/* Generated by re2c 1.3 */
#include <stdlib.h>
#include "chunk.h"
#include "scanners.h"

bufsize_t _scan_at(bufsize_t (*scanner)(const unsigned char *), cmark_chunk *c,
                   bufsize_t offset) {
  bufsize_t res;
  unsigned char *ptr = (unsigned char *)c->data;

  if (ptr == NULL || offset > c->len) {
    return 0;
  } else {
    unsigned char lim = ptr[c->len];

    ptr[c->len] = '\0';
    res = scanner(ptr + offset);
    ptr[c->len] = lim;
  }

  return res;
}

// Try to match a scheme including colon.
bufsize_t _scan_scheme(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    yych = *p;
    if (yych <= '@')
      goto yy2;
    if (yych <= 'Z')
      goto yy4;
    if (yych <= '`')
      goto yy2;
    if (yych <= 'z')
      goto yy4;
  yy2:
    ++p;
  yy3 : { return 0; }
  yy4:
    yych = *(marker = ++p);
    if (yych <= '/') {
      if (yych <= '+') {
        if (yych <= '*')
          goto yy3;
      } else {
        if (yych <= ',')
          goto yy3;
        if (yych >= '/')
          goto yy3;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '9')
          goto yy5;
        if (yych <= '@')
          goto yy3;
      } else {
        if (yych <= '`')
          goto yy3;
        if (yych >= '{')
          goto yy3;
      }
    }
  yy5:
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych == '+')
          goto yy7;
      } else {
        if (yych != '/')
          goto yy7;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych >= 'A')
          goto yy7;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych <= 'z')
          goto yy7;
      }
    }
  yy6:
    p = marker;
    goto yy3;
  yy7:
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych == '+')
          goto yy10;
        goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
        goto yy10;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
        goto yy10;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych <= 'z')
          goto yy10;
        goto yy6;
      }
    }
  yy8:
    ++p;
    { return (bufsize_t)(p - start); }
  yy10:
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy6;
      } else {
        if (yych == '/')
          goto yy6;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy8;
        if (yych <= '@')
          goto yy6;
      } else {
        if (yych <= '`')
          goto yy6;
        if (yych >= '{')
          goto yy6;
      }
    }
    yych = *++p;
    if (yych == ':')
      goto yy8;
    goto yy6;
  }
}

// Try to match URI autolink after first <, returning number of chars matched.
bufsize_t _scan_autolink_uri(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 0,   128, 0,   128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych <= '@')
      goto yy41;
    if (yych <= 'Z')
      goto yy43;
    if (yych <= '`')
      goto yy41;
    if (yych <= 'z')
      goto yy43;
  yy41:
    ++p;
  yy42 : { return 0; }
  yy43:
    yych = *(marker = ++p);
    if (yych <= '/') {
      if (yych <= '+') {
        if (yych <= '*')
          goto yy42;
      } else {
        if (yych <= ',')
          goto yy42;
        if (yych >= '/')
          goto yy42;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '9')
          goto yy44;
        if (yych <= '@')
          goto yy42;
      } else {
        if (yych <= '`')
          goto yy42;
        if (yych >= '{')
          goto yy42;
      }
    }
  yy44:
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych == '+')
          goto yy46;
      } else {
        if (yych != '/')
          goto yy46;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych >= 'A')
          goto yy46;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych <= 'z')
          goto yy46;
      }
    }
  yy45:
    p = marker;
    goto yy42;
  yy46:
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych == '+')
          goto yy49;
        goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
        goto yy49;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
        goto yy49;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych <= 'z')
          goto yy49;
        goto yy45;
      }
    }
  yy47:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy47;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '<')
          goto yy45;
        if (yych <= '>')
          goto yy50;
        goto yy45;
      } else {
        if (yych <= 0xDF)
          goto yy52;
        if (yych <= 0xE0)
          goto yy53;
        goto yy54;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy55;
        if (yych <= 0xEF)
          goto yy54;
        goto yy56;
      } else {
        if (yych <= 0xF3)
          goto yy57;
        if (yych <= 0xF4)
          goto yy58;
        goto yy45;
      }
    }
  yy49:
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych == '+')
          goto yy59;
        goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
        goto yy59;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
        goto yy59;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych <= 'z')
          goto yy59;
        goto yy45;
      }
    }
  yy50:
    ++p;
    { return (bufsize_t)(p - start); }
  yy52:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy45;
    if (yych <= 0xBF)
      goto yy47;
    goto yy45;
  yy53:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy45;
    if (yych <= 0xBF)
      goto yy52;
    goto yy45;
  yy54:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy45;
    if (yych <= 0xBF)
      goto yy52;
    goto yy45;
  yy55:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy45;
    if (yych <= 0x9F)
      goto yy52;
    goto yy45;
  yy56:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy45;
    if (yych <= 0xBF)
      goto yy54;
    goto yy45;
  yy57:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy45;
    if (yych <= 0xBF)
      goto yy54;
    goto yy45;
  yy58:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy45;
    if (yych <= 0x8F)
      goto yy54;
    goto yy45;
  yy59:
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych <= ',') {
        if (yych != '+')
          goto yy45;
      } else {
        if (yych == '/')
          goto yy45;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= ':')
          goto yy47;
        if (yych <= '@')
          goto yy45;
      } else {
        if (yych <= '`')
          goto yy45;
        if (yych >= '{')
          goto yy45;
      }
    }
    yych = *++p;
    if (yych == ':')
      goto yy47;
    goto yy45;
  }
}

// Try to match email autolink after first <, returning num of chars matched.
bufsize_t _scan_autolink_email(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   128, 0,   128, 128, 128, 128, 128, 0,   0,
        128, 128, 0,   128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 0,   0,   0,   128, 0,   128, 0,   128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 0,   0,   0,   128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych <= '9') {
      if (yych <= '\'') {
        if (yych == '!')
          goto yy91;
        if (yych >= '#')
          goto yy91;
      } else {
        if (yych <= ')')
          goto yy89;
        if (yych != ',')
          goto yy91;
      }
    } else {
      if (yych <= '?') {
        if (yych == '=')
          goto yy91;
        if (yych >= '?')
          goto yy91;
      } else {
        if (yych <= 'Z') {
          if (yych >= 'A')
            goto yy91;
        } else {
          if (yych <= ']')
            goto yy89;
          if (yych <= '~')
            goto yy91;
        }
      }
    }
  yy89:
    ++p;
  yy90 : { return 0; }
  yy91:
    yych = *(marker = ++p);
    if (yych <= ',') {
      if (yych <= '"') {
        if (yych == '!')
          goto yy93;
        goto yy90;
      } else {
        if (yych <= '\'')
          goto yy93;
        if (yych <= ')')
          goto yy90;
        if (yych <= '+')
          goto yy93;
        goto yy90;
      }
    } else {
      if (yych <= '>') {
        if (yych <= '9')
          goto yy93;
        if (yych == '=')
          goto yy93;
        goto yy90;
      } else {
        if (yych <= 'Z')
          goto yy93;
        if (yych <= ']')
          goto yy90;
        if (yych <= '~')
          goto yy93;
        goto yy90;
      }
    }
  yy92:
    yych = *++p;
  yy93:
    if (yybm[0 + yych] & 128) {
      goto yy92;
    }
    if (yych <= '>')
      goto yy94;
    if (yych <= '@')
      goto yy95;
  yy94:
    p = marker;
    goto yy90;
  yy95:
    yych = *++p;
    if (yych <= '@') {
      if (yych <= '/')
        goto yy94;
      if (yych >= ':')
        goto yy94;
    } else {
      if (yych <= 'Z')
        goto yy96;
      if (yych <= '`')
        goto yy94;
      if (yych >= '{')
        goto yy94;
    }
  yy96:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy98;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy98;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy98;
        goto yy94;
      }
    }
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy101;
      if (yych <= '/')
        goto yy94;
      goto yy102;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy102;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy102;
        goto yy94;
      }
    }
  yy98:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych <= '-')
          goto yy101;
        goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy102;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy102;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy102;
        goto yy94;
      }
    }
  yy99:
    ++p;
    { return (bufsize_t)(p - start); }
  yy101:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy103;
      if (yych <= '/')
        goto yy94;
      goto yy104;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy104;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy104;
        goto yy94;
      }
    }
  yy102:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy104;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy104;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy104;
        goto yy94;
      }
    }
  yy103:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy105;
      if (yych <= '/')
        goto yy94;
      goto yy106;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy106;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy106;
        goto yy94;
      }
    }
  yy104:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy106;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy106;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy106;
        goto yy94;
      }
    }
  yy105:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy107;
      if (yych <= '/')
        goto yy94;
      goto yy108;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy108;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy108;
        goto yy94;
      }
    }
  yy106:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy108;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy108;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy108;
        goto yy94;
      }
    }
  yy107:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy109;
      if (yych <= '/')
        goto yy94;
      goto yy110;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy110;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy110;
        goto yy94;
      }
    }
  yy108:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy110;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy110;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy110;
        goto yy94;
      }
    }
  yy109:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy111;
      if (yych <= '/')
        goto yy94;
      goto yy112;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy112;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy112;
        goto yy94;
      }
    }
  yy110:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy112;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy112;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy112;
        goto yy94;
      }
    }
  yy111:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy113;
      if (yych <= '/')
        goto yy94;
      goto yy114;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy114;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy114;
        goto yy94;
      }
    }
  yy112:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy114;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy114;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy114;
        goto yy94;
      }
    }
  yy113:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy115;
      if (yych <= '/')
        goto yy94;
      goto yy116;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy116;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy116;
        goto yy94;
      }
    }
  yy114:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy116;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy116;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy116;
        goto yy94;
      }
    }
  yy115:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy117;
      if (yych <= '/')
        goto yy94;
      goto yy118;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy118;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy118;
        goto yy94;
      }
    }
  yy116:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy118;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy118;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy118;
        goto yy94;
      }
    }
  yy117:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy119;
      if (yych <= '/')
        goto yy94;
      goto yy120;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy120;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy120;
        goto yy94;
      }
    }
  yy118:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy120;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy120;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy120;
        goto yy94;
      }
    }
  yy119:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy121;
      if (yych <= '/')
        goto yy94;
      goto yy122;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy122;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy122;
        goto yy94;
      }
    }
  yy120:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy122;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy122;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy122;
        goto yy94;
      }
    }
  yy121:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy123;
      if (yych <= '/')
        goto yy94;
      goto yy124;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy124;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy124;
        goto yy94;
      }
    }
  yy122:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy124;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy124;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy124;
        goto yy94;
      }
    }
  yy123:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy125;
      if (yych <= '/')
        goto yy94;
      goto yy126;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy126;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy126;
        goto yy94;
      }
    }
  yy124:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy126;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy126;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy126;
        goto yy94;
      }
    }
  yy125:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy127;
      if (yych <= '/')
        goto yy94;
      goto yy128;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy128;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy128;
        goto yy94;
      }
    }
  yy126:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy128;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy128;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy128;
        goto yy94;
      }
    }
  yy127:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy129;
      if (yych <= '/')
        goto yy94;
      goto yy130;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy130;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy130;
        goto yy94;
      }
    }
  yy128:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy130;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy130;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy130;
        goto yy94;
      }
    }
  yy129:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy131;
      if (yych <= '/')
        goto yy94;
      goto yy132;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy132;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy132;
        goto yy94;
      }
    }
  yy130:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy132;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy132;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy132;
        goto yy94;
      }
    }
  yy131:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy133;
      if (yych <= '/')
        goto yy94;
      goto yy134;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy134;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy134;
        goto yy94;
      }
    }
  yy132:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy134;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy134;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy134;
        goto yy94;
      }
    }
  yy133:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy135;
      if (yych <= '/')
        goto yy94;
      goto yy136;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy136;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy136;
        goto yy94;
      }
    }
  yy134:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy136;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy136;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy136;
        goto yy94;
      }
    }
  yy135:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy137;
      if (yych <= '/')
        goto yy94;
      goto yy138;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy138;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy138;
        goto yy94;
      }
    }
  yy136:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy138;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy138;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy138;
        goto yy94;
      }
    }
  yy137:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy139;
      if (yych <= '/')
        goto yy94;
      goto yy140;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy140;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy140;
        goto yy94;
      }
    }
  yy138:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy140;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy140;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy140;
        goto yy94;
      }
    }
  yy139:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy141;
      if (yych <= '/')
        goto yy94;
      goto yy142;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy142;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy142;
        goto yy94;
      }
    }
  yy140:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy142;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy142;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy142;
        goto yy94;
      }
    }
  yy141:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy143;
      if (yych <= '/')
        goto yy94;
      goto yy144;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy144;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy144;
        goto yy94;
      }
    }
  yy142:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy144;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy144;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy144;
        goto yy94;
      }
    }
  yy143:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy145;
      if (yych <= '/')
        goto yy94;
      goto yy146;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy146;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy146;
        goto yy94;
      }
    }
  yy144:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy146;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy146;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy146;
        goto yy94;
      }
    }
  yy145:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy147;
      if (yych <= '/')
        goto yy94;
      goto yy148;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy148;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy148;
        goto yy94;
      }
    }
  yy146:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy148;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy148;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy148;
        goto yy94;
      }
    }
  yy147:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy149;
      if (yych <= '/')
        goto yy94;
      goto yy150;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy150;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy150;
        goto yy94;
      }
    }
  yy148:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy150;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy150;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy150;
        goto yy94;
      }
    }
  yy149:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy151;
      if (yych <= '/')
        goto yy94;
      goto yy152;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy152;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy152;
        goto yy94;
      }
    }
  yy150:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy152;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy152;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy152;
        goto yy94;
      }
    }
  yy151:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy153;
      if (yych <= '/')
        goto yy94;
      goto yy154;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy154;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy154;
        goto yy94;
      }
    }
  yy152:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy154;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy154;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy154;
        goto yy94;
      }
    }
  yy153:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy155;
      if (yych <= '/')
        goto yy94;
      goto yy156;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy156;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy156;
        goto yy94;
      }
    }
  yy154:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy156;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy156;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy156;
        goto yy94;
      }
    }
  yy155:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy157;
      if (yych <= '/')
        goto yy94;
      goto yy158;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy158;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy158;
        goto yy94;
      }
    }
  yy156:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy158;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy158;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy158;
        goto yy94;
      }
    }
  yy157:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy159;
      if (yych <= '/')
        goto yy94;
      goto yy160;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy160;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy160;
        goto yy94;
      }
    }
  yy158:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy160;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy160;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy160;
        goto yy94;
      }
    }
  yy159:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy161;
      if (yych <= '/')
        goto yy94;
      goto yy162;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy162;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy162;
        goto yy94;
      }
    }
  yy160:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy162;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy162;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy162;
        goto yy94;
      }
    }
  yy161:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy163;
      if (yych <= '/')
        goto yy94;
      goto yy164;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy164;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy164;
        goto yy94;
      }
    }
  yy162:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy164;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy164;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy164;
        goto yy94;
      }
    }
  yy163:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy165;
      if (yych <= '/')
        goto yy94;
      goto yy166;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy166;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy166;
        goto yy94;
      }
    }
  yy164:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy166;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy166;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy166;
        goto yy94;
      }
    }
  yy165:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy167;
      if (yych <= '/')
        goto yy94;
      goto yy168;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy168;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy168;
        goto yy94;
      }
    }
  yy166:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy168;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy168;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy168;
        goto yy94;
      }
    }
  yy167:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy169;
      if (yych <= '/')
        goto yy94;
      goto yy170;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy170;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy170;
        goto yy94;
      }
    }
  yy168:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy170;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy170;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy170;
        goto yy94;
      }
    }
  yy169:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy171;
      if (yych <= '/')
        goto yy94;
      goto yy172;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy172;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy172;
        goto yy94;
      }
    }
  yy170:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy172;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy172;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy172;
        goto yy94;
      }
    }
  yy171:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy173;
      if (yych <= '/')
        goto yy94;
      goto yy174;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy174;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy174;
        goto yy94;
      }
    }
  yy172:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy174;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy174;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy174;
        goto yy94;
      }
    }
  yy173:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy175;
      if (yych <= '/')
        goto yy94;
      goto yy176;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy176;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy176;
        goto yy94;
      }
    }
  yy174:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy176;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy176;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy176;
        goto yy94;
      }
    }
  yy175:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy177;
      if (yych <= '/')
        goto yy94;
      goto yy178;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy178;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy178;
        goto yy94;
      }
    }
  yy176:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy178;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy178;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy178;
        goto yy94;
      }
    }
  yy177:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy179;
      if (yych <= '/')
        goto yy94;
      goto yy180;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy180;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy180;
        goto yy94;
      }
    }
  yy178:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy180;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy180;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy180;
        goto yy94;
      }
    }
  yy179:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy181;
      if (yych <= '/')
        goto yy94;
      goto yy182;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy182;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy182;
        goto yy94;
      }
    }
  yy180:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy182;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy182;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy182;
        goto yy94;
      }
    }
  yy181:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy183;
      if (yych <= '/')
        goto yy94;
      goto yy184;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy184;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy184;
        goto yy94;
      }
    }
  yy182:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy184;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy184;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy184;
        goto yy94;
      }
    }
  yy183:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy185;
      if (yych <= '/')
        goto yy94;
      goto yy186;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy186;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy186;
        goto yy94;
      }
    }
  yy184:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy186;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy186;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy186;
        goto yy94;
      }
    }
  yy185:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy187;
      if (yych <= '/')
        goto yy94;
      goto yy188;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy188;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy188;
        goto yy94;
      }
    }
  yy186:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy188;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy188;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy188;
        goto yy94;
      }
    }
  yy187:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy189;
      if (yych <= '/')
        goto yy94;
      goto yy190;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy190;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy190;
        goto yy94;
      }
    }
  yy188:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy190;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy190;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy190;
        goto yy94;
      }
    }
  yy189:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy191;
      if (yych <= '/')
        goto yy94;
      goto yy192;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy192;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy192;
        goto yy94;
      }
    }
  yy190:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy192;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy192;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy192;
        goto yy94;
      }
    }
  yy191:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy193;
      if (yych <= '/')
        goto yy94;
      goto yy194;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy194;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy194;
        goto yy94;
      }
    }
  yy192:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy194;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy194;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy194;
        goto yy94;
      }
    }
  yy193:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy195;
      if (yych <= '/')
        goto yy94;
      goto yy196;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy196;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy196;
        goto yy94;
      }
    }
  yy194:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy196;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy196;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy196;
        goto yy94;
      }
    }
  yy195:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy197;
      if (yych <= '/')
        goto yy94;
      goto yy198;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy198;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy198;
        goto yy94;
      }
    }
  yy196:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy198;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy198;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy198;
        goto yy94;
      }
    }
  yy197:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy199;
      if (yych <= '/')
        goto yy94;
      goto yy200;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy200;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy200;
        goto yy94;
      }
    }
  yy198:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy200;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy200;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy200;
        goto yy94;
      }
    }
  yy199:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy201;
      if (yych <= '/')
        goto yy94;
      goto yy202;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy202;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy202;
        goto yy94;
      }
    }
  yy200:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy202;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy202;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy202;
        goto yy94;
      }
    }
  yy201:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy203;
      if (yych <= '/')
        goto yy94;
      goto yy204;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy204;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy204;
        goto yy94;
      }
    }
  yy202:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy204;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy204;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy204;
        goto yy94;
      }
    }
  yy203:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy205;
      if (yych <= '/')
        goto yy94;
      goto yy206;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy206;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy206;
        goto yy94;
      }
    }
  yy204:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy206;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy206;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy206;
        goto yy94;
      }
    }
  yy205:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy207;
      if (yych <= '/')
        goto yy94;
      goto yy208;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy208;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy208;
        goto yy94;
      }
    }
  yy206:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy208;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy208;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy208;
        goto yy94;
      }
    }
  yy207:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy209;
      if (yych <= '/')
        goto yy94;
      goto yy210;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy210;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy210;
        goto yy94;
      }
    }
  yy208:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy210;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy210;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy210;
        goto yy94;
      }
    }
  yy209:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy211;
      if (yych <= '/')
        goto yy94;
      goto yy212;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy212;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy212;
        goto yy94;
      }
    }
  yy210:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy212;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy212;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy212;
        goto yy94;
      }
    }
  yy211:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy213;
      if (yych <= '/')
        goto yy94;
      goto yy214;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy214;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy214;
        goto yy94;
      }
    }
  yy212:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy214;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy214;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy214;
        goto yy94;
      }
    }
  yy213:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy215;
      if (yych <= '/')
        goto yy94;
      goto yy216;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy216;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy216;
        goto yy94;
      }
    }
  yy214:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy216;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy216;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy216;
        goto yy94;
      }
    }
  yy215:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy217;
      if (yych <= '/')
        goto yy94;
      goto yy218;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy218;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy218;
        goto yy94;
      }
    }
  yy216:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy218;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy218;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy218;
        goto yy94;
      }
    }
  yy217:
    yych = *++p;
    if (yych <= '9') {
      if (yych == '-')
        goto yy219;
      if (yych <= '/')
        goto yy94;
      goto yy220;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy94;
        goto yy220;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy220;
        goto yy94;
      }
    }
  yy218:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= ',')
          goto yy94;
        if (yych >= '.')
          goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych <= '9')
          goto yy220;
        goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
        goto yy220;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych <= 'z')
          goto yy220;
        goto yy94;
      }
    }
  yy219:
    yych = *++p;
    if (yych <= '@') {
      if (yych <= '/')
        goto yy94;
      if (yych <= '9')
        goto yy221;
      goto yy94;
    } else {
      if (yych <= 'Z')
        goto yy221;
      if (yych <= '`')
        goto yy94;
      if (yych <= 'z')
        goto yy221;
      goto yy94;
    }
  yy220:
    yych = *++p;
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych <= '-')
          goto yy94;
        goto yy95;
      } else {
        if (yych <= '/')
          goto yy94;
        if (yych >= ':')
          goto yy94;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy99;
        if (yych <= '@')
          goto yy94;
      } else {
        if (yych <= '`')
          goto yy94;
        if (yych >= '{')
          goto yy94;
      }
    }
  yy221:
    yych = *++p;
    if (yych == '.')
      goto yy95;
    if (yych == '>')
      goto yy99;
    goto yy94;
  }
}

// Try to match an HTML tag after first <, returning num of chars matched.
bufsize_t _scan_html_tag(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0,   224, 224, 224, 224, 224, 224, 224, 224, 200, 200, 200, 200, 200,
        224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224,
        224, 224, 224, 224, 200, 224, 128, 224, 224, 224, 224, 64,  224, 224,
        224, 224, 224, 244, 240, 224, 244, 244, 244, 244, 244, 244, 244, 244,
        244, 244, 240, 224, 192, 192, 192, 224, 224, 244, 244, 244, 244, 244,
        244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244,
        244, 244, 244, 244, 244, 244, 244, 224, 224, 224, 224, 240, 192, 244,
        244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244,
        244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 244, 224, 224, 224,
        224, 224, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych <= '@') {
      if (yych == '/')
        goto yy226;
    } else {
      if (yych <= 'Z')
        goto yy227;
      if (yych <= '`')
        goto yy224;
      if (yych <= 'z')
        goto yy227;
    }
  yy224:
    ++p;
  yy225 : { return 0; }
  yy226:
    yych = *(marker = ++p);
    if (yych <= '@')
      goto yy225;
    if (yych <= 'Z')
      goto yy228;
    if (yych <= '`')
      goto yy225;
    if (yych <= 'z')
      goto yy228;
    goto yy225;
  yy227:
    yych = *(marker = ++p);
    if (yych <= '.') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy225;
        if (yych <= '\r')
          goto yy234;
        goto yy225;
      } else {
        if (yych <= ' ')
          goto yy234;
        if (yych == '-')
          goto yy234;
        goto yy225;
      }
    } else {
      if (yych <= '@') {
        if (yych <= '9')
          goto yy234;
        if (yych == '>')
          goto yy234;
        goto yy225;
      } else {
        if (yych <= 'Z')
          goto yy234;
        if (yych <= '`')
          goto yy225;
        if (yych <= 'z')
          goto yy234;
        goto yy225;
      }
    }
  yy228:
    yych = *++p;
    if (yybm[0 + yych] & 4) {
      goto yy228;
    }
    if (yych <= 0x1F) {
      if (yych <= 0x08)
        goto yy230;
      if (yych <= '\r')
        goto yy238;
    } else {
      if (yych <= ' ')
        goto yy238;
      if (yych == '>')
        goto yy236;
    }
  yy230:
    p = marker;
    goto yy225;
  yy231:
    yych = *++p;
    if (yybm[0 + yych] & 8) {
      goto yy231;
    }
    if (yych <= '>') {
      if (yych <= '9') {
        if (yych == '/')
          goto yy235;
        goto yy230;
      } else {
        if (yych <= ':')
          goto yy240;
        if (yych <= '=')
          goto yy230;
        goto yy236;
      }
    } else {
      if (yych <= '^') {
        if (yych <= '@')
          goto yy230;
        if (yych <= 'Z')
          goto yy240;
        goto yy230;
      } else {
        if (yych == '`')
          goto yy230;
        if (yych <= 'z')
          goto yy240;
        goto yy230;
      }
    }
  yy233:
    yych = *++p;
  yy234:
    if (yybm[0 + yych] & 8) {
      goto yy231;
    }
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych == '-')
          goto yy233;
        goto yy230;
      } else {
        if (yych <= '/')
          goto yy235;
        if (yych <= '9')
          goto yy233;
        goto yy230;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy236;
        if (yych <= '@')
          goto yy230;
        goto yy233;
      } else {
        if (yych <= '`')
          goto yy230;
        if (yych <= 'z')
          goto yy233;
        goto yy230;
      }
    }
  yy235:
    yych = *++p;
    if (yych != '>')
      goto yy230;
  yy236:
    ++p;
    { return (bufsize_t)(p - start); }
  yy238:
    yych = *++p;
    if (yych <= 0x1F) {
      if (yych <= 0x08)
        goto yy230;
      if (yych <= '\r')
        goto yy238;
      goto yy230;
    } else {
      if (yych <= ' ')
        goto yy238;
      if (yych == '>')
        goto yy236;
      goto yy230;
    }
  yy240:
    yych = *++p;
    if (yybm[0 + yych] & 16) {
      goto yy240;
    }
    if (yych <= ',') {
      if (yych <= '\r') {
        if (yych <= 0x08)
          goto yy230;
      } else {
        if (yych != ' ')
          goto yy230;
      }
    } else {
      if (yych <= '<') {
        if (yych <= '/')
          goto yy235;
        goto yy230;
      } else {
        if (yych <= '=')
          goto yy244;
        if (yych <= '>')
          goto yy236;
        goto yy230;
      }
    }
  yy242:
    yych = *++p;
    if (yych <= '<') {
      if (yych <= ' ') {
        if (yych <= 0x08)
          goto yy230;
        if (yych <= '\r')
          goto yy242;
        if (yych <= 0x1F)
          goto yy230;
        goto yy242;
      } else {
        if (yych <= '/') {
          if (yych <= '.')
            goto yy230;
          goto yy235;
        } else {
          if (yych == ':')
            goto yy240;
          goto yy230;
        }
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '=')
          goto yy244;
        if (yych <= '>')
          goto yy236;
        if (yych <= '@')
          goto yy230;
        goto yy240;
      } else {
        if (yych <= '_') {
          if (yych <= '^')
            goto yy230;
          goto yy240;
        } else {
          if (yych <= '`')
            goto yy230;
          if (yych <= 'z')
            goto yy240;
          goto yy230;
        }
      }
    }
  yy244:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy246;
    }
    if (yych <= 0xE0) {
      if (yych <= '"') {
        if (yych <= 0x00)
          goto yy230;
        if (yych <= ' ')
          goto yy244;
        goto yy248;
      } else {
        if (yych <= '\'')
          goto yy250;
        if (yych <= 0xC1)
          goto yy230;
        if (yych <= 0xDF)
          goto yy252;
        goto yy253;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy255;
        goto yy254;
      } else {
        if (yych <= 0xF0)
          goto yy256;
        if (yych <= 0xF3)
          goto yy257;
        if (yych <= 0xF4)
          goto yy258;
        goto yy230;
      }
    }
  yy246:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy246;
    }
    if (yych <= 0xE0) {
      if (yych <= '=') {
        if (yych <= 0x00)
          goto yy230;
        if (yych <= ' ')
          goto yy231;
        goto yy230;
      } else {
        if (yych <= '>')
          goto yy236;
        if (yych <= 0xC1)
          goto yy230;
        if (yych <= 0xDF)
          goto yy252;
        goto yy253;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy255;
        goto yy254;
      } else {
        if (yych <= 0xF0)
          goto yy256;
        if (yych <= 0xF3)
          goto yy257;
        if (yych <= 0xF4)
          goto yy258;
        goto yy230;
      }
    }
  yy248:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy248;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy230;
        if (yych <= '"')
          goto yy259;
        goto yy230;
      } else {
        if (yych <= 0xDF)
          goto yy260;
        if (yych <= 0xE0)
          goto yy261;
        goto yy262;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy263;
        if (yych <= 0xEF)
          goto yy262;
        goto yy264;
      } else {
        if (yych <= 0xF3)
          goto yy265;
        if (yych <= 0xF4)
          goto yy266;
        goto yy230;
      }
    }
  yy250:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy250;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy230;
        if (yych <= '\'')
          goto yy259;
        goto yy230;
      } else {
        if (yych <= 0xDF)
          goto yy267;
        if (yych <= 0xE0)
          goto yy268;
        goto yy269;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy270;
        if (yych <= 0xEF)
          goto yy269;
        goto yy271;
      } else {
        if (yych <= 0xF3)
          goto yy272;
        if (yych <= 0xF4)
          goto yy273;
        goto yy230;
      }
    }
  yy252:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy246;
    goto yy230;
  yy253:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy252;
    goto yy230;
  yy254:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy252;
    goto yy230;
  yy255:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0x9F)
      goto yy252;
    goto yy230;
  yy256:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy254;
    goto yy230;
  yy257:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy254;
    goto yy230;
  yy258:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0x8F)
      goto yy254;
    goto yy230;
  yy259:
    yych = *++p;
    if (yybm[0 + yych] & 8) {
      goto yy231;
    }
    if (yych == '/')
      goto yy235;
    if (yych == '>')
      goto yy236;
    goto yy230;
  yy260:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy248;
    goto yy230;
  yy261:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy260;
    goto yy230;
  yy262:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy260;
    goto yy230;
  yy263:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0x9F)
      goto yy260;
    goto yy230;
  yy264:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy262;
    goto yy230;
  yy265:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy262;
    goto yy230;
  yy266:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0x8F)
      goto yy262;
    goto yy230;
  yy267:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy250;
    goto yy230;
  yy268:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy267;
    goto yy230;
  yy269:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy267;
    goto yy230;
  yy270:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0x9F)
      goto yy267;
    goto yy230;
  yy271:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy269;
    goto yy230;
  yy272:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0xBF)
      goto yy269;
    goto yy230;
  yy273:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy230;
    if (yych <= 0x8F)
      goto yy269;
    goto yy230;
  }
}

bufsize_t _scan_html_comment(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0,   128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 0,   128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych == '-')
      goto yy278;
    ++p;
  yy277 : { return 0; }
  yy278:
    yych = *(marker = ++p);
    if (yych <= '=') {
      if (yych <= 0x00)
        goto yy277;
      if (yych == '-')
        goto yy282;
      goto yy280;
    } else {
      if (yych <= 0x7F) {
        if (yych <= '>')
          goto yy277;
        goto yy280;
      } else {
        if (yych <= 0xC1)
          goto yy277;
        if (yych <= 0xF4)
          goto yy280;
        goto yy277;
      }
    }
  yy279:
    yych = *++p;
  yy280:
    if (yybm[0 + yych] & 128) {
      goto yy279;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy281;
        if (yych <= '-')
          goto yy290;
      } else {
        if (yych <= 0xDF)
          goto yy283;
        if (yych <= 0xE0)
          goto yy284;
        goto yy285;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy286;
        if (yych <= 0xEF)
          goto yy285;
        goto yy287;
      } else {
        if (yych <= 0xF3)
          goto yy288;
        if (yych <= 0xF4)
          goto yy289;
      }
    }
  yy281:
    p = marker;
    goto yy277;
  yy282:
    yych = *++p;
    if (yych == '-')
      goto yy291;
    if (yych == '>')
      goto yy281;
    goto yy280;
  yy283:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy281;
    if (yych <= 0xBF)
      goto yy279;
    goto yy281;
  yy284:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy281;
    if (yych <= 0xBF)
      goto yy283;
    goto yy281;
  yy285:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy281;
    if (yych <= 0xBF)
      goto yy283;
    goto yy281;
  yy286:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy281;
    if (yych <= 0x9F)
      goto yy283;
    goto yy281;
  yy287:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy281;
    if (yych <= 0xBF)
      goto yy285;
    goto yy281;
  yy288:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy281;
    if (yych <= 0xBF)
      goto yy285;
    goto yy281;
  yy289:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy281;
    if (yych <= 0x8F)
      goto yy285;
    goto yy281;
  yy290:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy279;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy281;
        if (yych >= '.')
          goto yy281;
      } else {
        if (yych <= 0xDF)
          goto yy283;
        if (yych <= 0xE0)
          goto yy284;
        goto yy285;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy286;
        if (yych <= 0xEF)
          goto yy285;
        goto yy287;
      } else {
        if (yych <= 0xF3)
          goto yy288;
        if (yych <= 0xF4)
          goto yy289;
        goto yy281;
      }
    }
  yy291:
    yych = *++p;
    if (yych != '>')
      goto yy281;
    ++p;
    { return (bufsize_t)(p - start); }
  }
}

bufsize_t _scan_html_pi(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,   128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 0,   128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yybm[0 + yych] & 128) {
      goto yy298;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy296;
        if (yych <= '?')
          goto yy301;
      } else {
        if (yych <= 0xDF)
          goto yy302;
        if (yych <= 0xE0)
          goto yy303;
        goto yy304;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy305;
        if (yych <= 0xEF)
          goto yy304;
        goto yy306;
      } else {
        if (yych <= 0xF3)
          goto yy307;
        if (yych <= 0xF4)
          goto yy308;
      }
    }
  yy296:
    ++p;
  yy297 : { return 0; }
  yy298:
    yyaccept = 0;
    yych = *(marker = ++p);
  yy299:
    if (yybm[0 + yych] & 128) {
      goto yy298;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy300;
        if (yych <= '?')
          goto yy309;
      } else {
        if (yych <= 0xDF)
          goto yy311;
        if (yych <= 0xE0)
          goto yy312;
        goto yy313;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy314;
        if (yych <= 0xEF)
          goto yy313;
        goto yy315;
      } else {
        if (yych <= 0xF3)
          goto yy316;
        if (yych <= 0xF4)
          goto yy317;
      }
    }
  yy300 : { return (bufsize_t)(p - start); }
  yy301:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych <= '?') {
      if (yych <= 0x00)
        goto yy297;
      if (yych <= '=')
        goto yy299;
      if (yych <= '>')
        goto yy297;
      goto yy298;
    } else {
      if (yych <= 0x7F)
        goto yy299;
      if (yych <= 0xC1)
        goto yy297;
      if (yych <= 0xF4)
        goto yy299;
      goto yy297;
    }
  yy302:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy297;
    if (yych <= 0xBF)
      goto yy298;
    goto yy297;
  yy303:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych <= 0x9F)
      goto yy297;
    if (yych <= 0xBF)
      goto yy311;
    goto yy297;
  yy304:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy297;
    if (yych <= 0xBF)
      goto yy311;
    goto yy297;
  yy305:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy297;
    if (yych <= 0x9F)
      goto yy311;
    goto yy297;
  yy306:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych <= 0x8F)
      goto yy297;
    if (yych <= 0xBF)
      goto yy313;
    goto yy297;
  yy307:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy297;
    if (yych <= 0xBF)
      goto yy313;
    goto yy297;
  yy308:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy297;
    if (yych <= 0x8F)
      goto yy313;
    goto yy297;
  yy309:
    yych = *++p;
    if (yych <= 0xE0) {
      if (yych <= '>') {
        if (yych <= 0x00)
          goto yy310;
        if (yych <= '=')
          goto yy298;
      } else {
        if (yych <= 0x7F)
          goto yy298;
        if (yych <= 0xC1)
          goto yy310;
        if (yych <= 0xDF)
          goto yy311;
        goto yy312;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy314;
        goto yy313;
      } else {
        if (yych <= 0xF0)
          goto yy315;
        if (yych <= 0xF3)
          goto yy316;
        if (yych <= 0xF4)
          goto yy317;
      }
    }
  yy310:
    p = marker;
    if (yyaccept == 0) {
      goto yy300;
    } else {
      goto yy297;
    }
  yy311:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy310;
    if (yych <= 0xBF)
      goto yy298;
    goto yy310;
  yy312:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy310;
    if (yych <= 0xBF)
      goto yy311;
    goto yy310;
  yy313:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy310;
    if (yych <= 0xBF)
      goto yy311;
    goto yy310;
  yy314:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy310;
    if (yych <= 0x9F)
      goto yy311;
    goto yy310;
  yy315:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy310;
    if (yych <= 0xBF)
      goto yy313;
    goto yy310;
  yy316:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy310;
    if (yych <= 0xBF)
      goto yy313;
    goto yy310;
  yy317:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy310;
    if (yych <= 0x8F)
      goto yy313;
    goto yy310;
  }
}

bufsize_t _scan_html_declaration(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,   64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,
        64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,
        64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,
        64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,
        64,  64,  64,  64,  64,  64,  0,   64,  64,  192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 64,  64,  64,  64,  64,  64,  64,
        64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,
        64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,  64,
        64,  64,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych <= '@')
      goto yy320;
    if (yych <= 'Z')
      goto yy322;
  yy320:
    ++p;
  yy321 : { return 0; }
  yy322:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy326;
    }
    if (yych <= 0x08)
      goto yy321;
    if (yych <= '\r')
      goto yy323;
    if (yych != ' ')
      goto yy321;
  yy323:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 64) {
      goto yy323;
    }
    if (yych <= 0xED) {
      if (yych <= 0xDF) {
        if (yych >= 0xC2)
          goto yy329;
      } else {
        if (yych <= 0xE0)
          goto yy330;
        if (yych <= 0xEC)
          goto yy331;
        goto yy332;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xEF)
          goto yy331;
        goto yy333;
      } else {
        if (yych <= 0xF3)
          goto yy334;
        if (yych <= 0xF4)
          goto yy335;
      }
    }
  yy325 : { return (bufsize_t)(p - start); }
  yy326:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy326;
    }
    if (yych <= 0x08)
      goto yy328;
    if (yych <= '\r')
      goto yy323;
    if (yych == ' ')
      goto yy323;
  yy328:
    p = marker;
    if (yyaccept == 0) {
      goto yy321;
    } else {
      goto yy325;
    }
  yy329:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy328;
    if (yych <= 0xBF)
      goto yy323;
    goto yy328;
  yy330:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy328;
    if (yych <= 0xBF)
      goto yy329;
    goto yy328;
  yy331:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy328;
    if (yych <= 0xBF)
      goto yy329;
    goto yy328;
  yy332:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy328;
    if (yych <= 0x9F)
      goto yy329;
    goto yy328;
  yy333:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy328;
    if (yych <= 0xBF)
      goto yy331;
    goto yy328;
  yy334:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy328;
    if (yych <= 0xBF)
      goto yy331;
    goto yy328;
  yy335:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy328;
    if (yych <= 0x8F)
      goto yy331;
    goto yy328;
  }
}

bufsize_t _scan_html_cdata(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,   128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 0,   128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych == 'C')
      goto yy340;
    if (yych == 'c')
      goto yy340;
    ++p;
  yy339 : { return 0; }
  yy340:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych == 'D')
      goto yy341;
    if (yych != 'd')
      goto yy339;
  yy341:
    yych = *++p;
    if (yych == 'A')
      goto yy343;
    if (yych == 'a')
      goto yy343;
  yy342:
    p = marker;
    if (yyaccept == 0) {
      goto yy339;
    } else {
      goto yy348;
    }
  yy343:
    yych = *++p;
    if (yych == 'T')
      goto yy344;
    if (yych != 't')
      goto yy342;
  yy344:
    yych = *++p;
    if (yych == 'A')
      goto yy345;
    if (yych != 'a')
      goto yy342;
  yy345:
    yych = *++p;
    if (yych != '[')
      goto yy342;
  yy346:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy346;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy348;
        if (yych <= ']')
          goto yy349;
      } else {
        if (yych <= 0xDF)
          goto yy350;
        if (yych <= 0xE0)
          goto yy351;
        goto yy352;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy353;
        if (yych <= 0xEF)
          goto yy352;
        goto yy354;
      } else {
        if (yych <= 0xF3)
          goto yy355;
        if (yych <= 0xF4)
          goto yy356;
      }
    }
  yy348 : { return (bufsize_t)(p - start); }
  yy349:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy346;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy342;
        if (yych <= ']')
          goto yy357;
        goto yy342;
      } else {
        if (yych <= 0xDF)
          goto yy350;
        if (yych <= 0xE0)
          goto yy351;
        goto yy352;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy353;
        if (yych <= 0xEF)
          goto yy352;
        goto yy354;
      } else {
        if (yych <= 0xF3)
          goto yy355;
        if (yych <= 0xF4)
          goto yy356;
        goto yy342;
      }
    }
  yy350:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy342;
    if (yych <= 0xBF)
      goto yy346;
    goto yy342;
  yy351:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy342;
    if (yych <= 0xBF)
      goto yy350;
    goto yy342;
  yy352:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy342;
    if (yych <= 0xBF)
      goto yy350;
    goto yy342;
  yy353:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy342;
    if (yych <= 0x9F)
      goto yy350;
    goto yy342;
  yy354:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy342;
    if (yych <= 0xBF)
      goto yy352;
    goto yy342;
  yy355:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy342;
    if (yych <= 0xBF)
      goto yy352;
    goto yy342;
  yy356:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy342;
    if (yych <= 0x8F)
      goto yy352;
    goto yy342;
  yy357:
    yych = *++p;
    if (yych <= 0xE0) {
      if (yych <= '>') {
        if (yych <= 0x00)
          goto yy342;
        if (yych <= '=')
          goto yy346;
        goto yy342;
      } else {
        if (yych <= 0x7F)
          goto yy346;
        if (yych <= 0xC1)
          goto yy342;
        if (yych <= 0xDF)
          goto yy350;
        goto yy351;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy353;
        goto yy352;
      } else {
        if (yych <= 0xF0)
          goto yy354;
        if (yych <= 0xF3)
          goto yy355;
        if (yych <= 0xF4)
          goto yy356;
        goto yy342;
      }
    }
  }
}

// Try to match an HTML block tag start line, returning
// an integer code for the type of block (1-6, matching the spec).
// #7 is handled by a separate function, below.
bufsize_t _scan_html_block_start(const unsigned char *p) {
  const unsigned char *marker = NULL;

  {
    unsigned char yych;
    yych = *p;
    if (yych == '<')
      goto yy362;
    ++p;
  yy361 : { return 0; }
  yy362:
    yych = *(marker = ++p);
    switch (yych) {
    case '!':
      goto yy363;
    case '/':
      goto yy365;
    case '?':
      goto yy366;
    case 'A':
    case 'a':
      goto yy368;
    case 'B':
    case 'b':
      goto yy369;
    case 'C':
    case 'c':
      goto yy370;
    case 'D':
    case 'd':
      goto yy371;
    case 'F':
    case 'f':
      goto yy372;
    case 'H':
    case 'h':
      goto yy373;
    case 'I':
    case 'i':
      goto yy374;
    case 'L':
    case 'l':
      goto yy375;
    case 'M':
    case 'm':
      goto yy376;
    case 'N':
    case 'n':
      goto yy377;
    case 'O':
    case 'o':
      goto yy378;
    case 'P':
    case 'p':
      goto yy379;
    case 'S':
    case 's':
      goto yy380;
    case 'T':
    case 't':
      goto yy381;
    case 'U':
    case 'u':
      goto yy382;
    default:
      goto yy361;
    }
  yy363:
    yych = *++p;
    if (yych <= '@') {
      if (yych == '-')
        goto yy383;
    } else {
      if (yych <= 'Z')
        goto yy384;
      if (yych <= '[')
        goto yy386;
    }
  yy364:
    p = marker;
    goto yy361;
  yy365:
    yych = *++p;
    switch (yych) {
    case 'A':
    case 'a':
      goto yy368;
    case 'B':
    case 'b':
      goto yy369;
    case 'C':
    case 'c':
      goto yy370;
    case 'D':
    case 'd':
      goto yy371;
    case 'F':
    case 'f':
      goto yy372;
    case 'H':
    case 'h':
      goto yy373;
    case 'I':
    case 'i':
      goto yy374;
    case 'L':
    case 'l':
      goto yy375;
    case 'M':
    case 'm':
      goto yy376;
    case 'N':
    case 'n':
      goto yy377;
    case 'O':
    case 'o':
      goto yy378;
    case 'P':
    case 'p':
      goto yy387;
    case 'S':
    case 's':
      goto yy388;
    case 'T':
    case 't':
      goto yy389;
    case 'U':
    case 'u':
      goto yy382;
    default:
      goto yy364;
    }
  yy366:
    ++p;
    { return 3; }
  yy368:
    yych = *++p;
    if (yych <= 'S') {
      if (yych <= 'D') {
        if (yych <= 'C')
          goto yy364;
        goto yy390;
      } else {
        if (yych <= 'Q')
          goto yy364;
        if (yych <= 'R')
          goto yy391;
        goto yy392;
      }
    } else {
      if (yych <= 'q') {
        if (yych == 'd')
          goto yy390;
        goto yy364;
      } else {
        if (yych <= 'r')
          goto yy391;
        if (yych <= 's')
          goto yy392;
        goto yy364;
      }
    }
  yy369:
    yych = *++p;
    if (yych <= 'O') {
      if (yych <= 'K') {
        if (yych == 'A')
          goto yy393;
        goto yy364;
      } else {
        if (yych <= 'L')
          goto yy394;
        if (yych <= 'N')
          goto yy364;
        goto yy395;
      }
    } else {
      if (yych <= 'k') {
        if (yych == 'a')
          goto yy393;
        goto yy364;
      } else {
        if (yych <= 'l')
          goto yy394;
        if (yych == 'o')
          goto yy395;
        goto yy364;
      }
    }
  yy370:
    yych = *++p;
    if (yych <= 'O') {
      if (yych <= 'D') {
        if (yych == 'A')
          goto yy396;
        goto yy364;
      } else {
        if (yych <= 'E')
          goto yy397;
        if (yych <= 'N')
          goto yy364;
        goto yy398;
      }
    } else {
      if (yych <= 'd') {
        if (yych == 'a')
          goto yy396;
        goto yy364;
      } else {
        if (yych <= 'e')
          goto yy397;
        if (yych == 'o')
          goto yy398;
        goto yy364;
      }
    }
  yy371:
    yych = *++p;
    switch (yych) {
    case 'D':
    case 'L':
    case 'T':
    case 'd':
    case 'l':
    case 't':
      goto yy399;
    case 'E':
    case 'e':
      goto yy400;
    case 'I':
    case 'i':
      goto yy401;
    default:
      goto yy364;
    }
  yy372:
    yych = *++p;
    if (yych <= 'R') {
      if (yych <= 'N') {
        if (yych == 'I')
          goto yy402;
        goto yy364;
      } else {
        if (yych <= 'O')
          goto yy403;
        if (yych <= 'Q')
          goto yy364;
        goto yy404;
      }
    } else {
      if (yych <= 'n') {
        if (yych == 'i')
          goto yy402;
        goto yy364;
      } else {
        if (yych <= 'o')
          goto yy403;
        if (yych == 'r')
          goto yy404;
        goto yy364;
      }
    }
  yy373:
    yych = *++p;
    if (yych <= 'S') {
      if (yych <= 'D') {
        if (yych <= '0')
          goto yy364;
        if (yych <= '6')
          goto yy399;
        goto yy364;
      } else {
        if (yych <= 'E')
          goto yy405;
        if (yych == 'R')
          goto yy399;
        goto yy364;
      }
    } else {
      if (yych <= 'q') {
        if (yych <= 'T')
          goto yy406;
        if (yych == 'e')
          goto yy405;
        goto yy364;
      } else {
        if (yych <= 'r')
          goto yy399;
        if (yych == 't')
          goto yy406;
        goto yy364;
      }
    }
  yy374:
    yych = *++p;
    if (yych == 'F')
      goto yy407;
    if (yych == 'f')
      goto yy407;
    goto yy364;
  yy375:
    yych = *++p;
    if (yych <= 'I') {
      if (yych == 'E')
        goto yy408;
      if (yych <= 'H')
        goto yy364;
      goto yy409;
    } else {
      if (yych <= 'e') {
        if (yych <= 'd')
          goto yy364;
        goto yy408;
      } else {
        if (yych == 'i')
          goto yy409;
        goto yy364;
      }
    }
  yy376:
    yych = *++p;
    if (yych <= 'E') {
      if (yych == 'A')
        goto yy410;
      if (yych <= 'D')
        goto yy364;
      goto yy411;
    } else {
      if (yych <= 'a') {
        if (yych <= '`')
          goto yy364;
        goto yy410;
      } else {
        if (yych == 'e')
          goto yy411;
        goto yy364;
      }
    }
  yy377:
    yych = *++p;
    if (yych <= 'O') {
      if (yych == 'A')
        goto yy412;
      if (yych <= 'N')
        goto yy364;
      goto yy413;
    } else {
      if (yych <= 'a') {
        if (yych <= '`')
          goto yy364;
        goto yy412;
      } else {
        if (yych == 'o')
          goto yy413;
        goto yy364;
      }
    }
  yy378:
    yych = *++p;
    if (yych <= 'P') {
      if (yych == 'L')
        goto yy399;
      if (yych <= 'O')
        goto yy364;
      goto yy414;
    } else {
      if (yych <= 'l') {
        if (yych <= 'k')
          goto yy364;
        goto yy399;
      } else {
        if (yych == 'p')
          goto yy414;
        goto yy364;
      }
    }
  yy379:
    yych = *++p;
    if (yych <= '>') {
      if (yych <= ' ') {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        if (yych <= 0x1F)
          goto yy364;
        goto yy415;
      } else {
        if (yych == '/')
          goto yy417;
        if (yych <= '=')
          goto yy364;
        goto yy415;
      }
    } else {
      if (yych <= 'R') {
        if (yych == 'A')
          goto yy418;
        if (yych <= 'Q')
          goto yy364;
        goto yy419;
      } else {
        if (yych <= 'a') {
          if (yych <= '`')
            goto yy364;
          goto yy418;
        } else {
          if (yych == 'r')
            goto yy419;
          goto yy364;
        }
      }
    }
  yy380:
    yych = *++p;
    switch (yych) {
    case 'C':
    case 'c':
      goto yy420;
    case 'E':
    case 'e':
      goto yy421;
    case 'O':
    case 'o':
      goto yy422;
    case 'T':
    case 't':
      goto yy423;
    case 'U':
    case 'u':
      goto yy424;
    default:
      goto yy364;
    }
  yy381:
    yych = *++p;
    switch (yych) {
    case 'A':
    case 'a':
      goto yy425;
    case 'B':
    case 'b':
      goto yy426;
    case 'D':
    case 'd':
      goto yy399;
    case 'E':
    case 'e':
      goto yy427;
    case 'F':
    case 'f':
      goto yy428;
    case 'H':
    case 'h':
      goto yy429;
    case 'I':
    case 'i':
      goto yy430;
    case 'R':
    case 'r':
      goto yy431;
    default:
      goto yy364;
    }
  yy382:
    yych = *++p;
    if (yych == 'L')
      goto yy399;
    if (yych == 'l')
      goto yy399;
    goto yy364;
  yy383:
    yych = *++p;
    if (yych == '-')
      goto yy432;
    goto yy364;
  yy384:
    ++p;
    { return 4; }
  yy386:
    yych = *++p;
    if (yych == 'C')
      goto yy434;
    if (yych == 'c')
      goto yy434;
    goto yy364;
  yy387:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= '@') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'A')
          goto yy418;
        if (yych == 'a')
          goto yy418;
        goto yy364;
      }
    }
  yy388:
    yych = *++p;
    if (yych <= 'U') {
      if (yych <= 'N') {
        if (yych == 'E')
          goto yy421;
        goto yy364;
      } else {
        if (yych <= 'O')
          goto yy422;
        if (yych <= 'T')
          goto yy364;
        goto yy424;
      }
    } else {
      if (yych <= 'n') {
        if (yych == 'e')
          goto yy421;
        goto yy364;
      } else {
        if (yych <= 'o')
          goto yy422;
        if (yych == 'u')
          goto yy424;
        goto yy364;
      }
    }
  yy389:
    yych = *++p;
    switch (yych) {
    case 'A':
    case 'a':
      goto yy425;
    case 'B':
    case 'b':
      goto yy426;
    case 'D':
    case 'd':
      goto yy399;
    case 'F':
    case 'f':
      goto yy428;
    case 'H':
    case 'h':
      goto yy429;
    case 'I':
    case 'i':
      goto yy430;
    case 'R':
    case 'r':
      goto yy431;
    default:
      goto yy364;
    }
  yy390:
    yych = *++p;
    if (yych == 'D')
      goto yy435;
    if (yych == 'd')
      goto yy435;
    goto yy364;
  yy391:
    yych = *++p;
    if (yych == 'T')
      goto yy436;
    if (yych == 't')
      goto yy436;
    goto yy364;
  yy392:
    yych = *++p;
    if (yych == 'I')
      goto yy437;
    if (yych == 'i')
      goto yy437;
    goto yy364;
  yy393:
    yych = *++p;
    if (yych == 'S')
      goto yy438;
    if (yych == 's')
      goto yy438;
    goto yy364;
  yy394:
    yych = *++p;
    if (yych == 'O')
      goto yy439;
    if (yych == 'o')
      goto yy439;
    goto yy364;
  yy395:
    yych = *++p;
    if (yych == 'D')
      goto yy440;
    if (yych == 'd')
      goto yy440;
    goto yy364;
  yy396:
    yych = *++p;
    if (yych == 'P')
      goto yy441;
    if (yych == 'p')
      goto yy441;
    goto yy364;
  yy397:
    yych = *++p;
    if (yych == 'N')
      goto yy442;
    if (yych == 'n')
      goto yy442;
    goto yy364;
  yy398:
    yych = *++p;
    if (yych == 'L')
      goto yy443;
    if (yych == 'l')
      goto yy443;
    goto yy364;
  yy399:
    yych = *++p;
    if (yych <= ' ') {
      if (yych <= 0x08)
        goto yy364;
      if (yych <= '\r')
        goto yy415;
      if (yych <= 0x1F)
        goto yy364;
      goto yy415;
    } else {
      if (yych <= '/') {
        if (yych <= '.')
          goto yy364;
        goto yy417;
      } else {
        if (yych == '>')
          goto yy415;
        goto yy364;
      }
    }
  yy400:
    yych = *++p;
    if (yych == 'T')
      goto yy444;
    if (yych == 't')
      goto yy444;
    goto yy364;
  yy401:
    yych = *++p;
    if (yych <= 'V') {
      if (yych <= 'Q') {
        if (yych == 'A')
          goto yy445;
        goto yy364;
      } else {
        if (yych <= 'R')
          goto yy399;
        if (yych <= 'U')
          goto yy364;
        goto yy399;
      }
    } else {
      if (yych <= 'q') {
        if (yych == 'a')
          goto yy445;
        goto yy364;
      } else {
        if (yych <= 'r')
          goto yy399;
        if (yych == 'v')
          goto yy399;
        goto yy364;
      }
    }
  yy402:
    yych = *++p;
    if (yych <= 'G') {
      if (yych == 'E')
        goto yy446;
      if (yych <= 'F')
        goto yy364;
      goto yy447;
    } else {
      if (yych <= 'e') {
        if (yych <= 'd')
          goto yy364;
        goto yy446;
      } else {
        if (yych == 'g')
          goto yy447;
        goto yy364;
      }
    }
  yy403:
    yych = *++p;
    if (yych <= 'R') {
      if (yych == 'O')
        goto yy442;
      if (yych <= 'Q')
        goto yy364;
      goto yy448;
    } else {
      if (yych <= 'o') {
        if (yych <= 'n')
          goto yy364;
        goto yy442;
      } else {
        if (yych == 'r')
          goto yy448;
        goto yy364;
      }
    }
  yy404:
    yych = *++p;
    if (yych == 'A')
      goto yy449;
    if (yych == 'a')
      goto yy449;
    goto yy364;
  yy405:
    yych = *++p;
    if (yych == 'A')
      goto yy450;
    if (yych == 'a')
      goto yy450;
    goto yy364;
  yy406:
    yych = *++p;
    if (yych == 'M')
      goto yy382;
    if (yych == 'm')
      goto yy382;
    goto yy364;
  yy407:
    yych = *++p;
    if (yych == 'R')
      goto yy451;
    if (yych == 'r')
      goto yy451;
    goto yy364;
  yy408:
    yych = *++p;
    if (yych == 'G')
      goto yy452;
    if (yych == 'g')
      goto yy452;
    goto yy364;
  yy409:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= 'M') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'N')
          goto yy453;
        if (yych == 'n')
          goto yy453;
        goto yy364;
      }
    }
  yy410:
    yych = *++p;
    if (yych == 'I')
      goto yy454;
    if (yych == 'i')
      goto yy454;
    goto yy364;
  yy411:
    yych = *++p;
    if (yych == 'N')
      goto yy455;
    if (yych == 'n')
      goto yy455;
    goto yy364;
  yy412:
    yych = *++p;
    if (yych == 'V')
      goto yy399;
    if (yych == 'v')
      goto yy399;
    goto yy364;
  yy413:
    yych = *++p;
    if (yych == 'F')
      goto yy456;
    if (yych == 'f')
      goto yy456;
    goto yy364;
  yy414:
    yych = *++p;
    if (yych == 'T')
      goto yy457;
    if (yych == 't')
      goto yy457;
    goto yy364;
  yy415:
    ++p;
    { return 6; }
  yy417:
    yych = *++p;
    if (yych == '>')
      goto yy415;
    goto yy364;
  yy418:
    yych = *++p;
    if (yych == 'R')
      goto yy458;
    if (yych == 'r')
      goto yy458;
    goto yy364;
  yy419:
    yych = *++p;
    if (yych == 'E')
      goto yy459;
    if (yych == 'e')
      goto yy459;
    goto yy364;
  yy420:
    yych = *++p;
    if (yych == 'R')
      goto yy460;
    if (yych == 'r')
      goto yy460;
    goto yy364;
  yy421:
    yych = *++p;
    if (yych == 'C')
      goto yy441;
    if (yych == 'c')
      goto yy441;
    goto yy364;
  yy422:
    yych = *++p;
    if (yych == 'U')
      goto yy461;
    if (yych == 'u')
      goto yy461;
    goto yy364;
  yy423:
    yych = *++p;
    if (yych == 'Y')
      goto yy462;
    if (yych == 'y')
      goto yy462;
    goto yy364;
  yy424:
    yych = *++p;
    if (yych == 'M')
      goto yy463;
    if (yych == 'm')
      goto yy463;
    goto yy364;
  yy425:
    yych = *++p;
    if (yych == 'B')
      goto yy464;
    if (yych == 'b')
      goto yy464;
    goto yy364;
  yy426:
    yych = *++p;
    if (yych == 'O')
      goto yy395;
    if (yych == 'o')
      goto yy395;
    goto yy364;
  yy427:
    yych = *++p;
    if (yych == 'X')
      goto yy465;
    if (yych == 'x')
      goto yy465;
    goto yy364;
  yy428:
    yych = *++p;
    if (yych == 'O')
      goto yy466;
    if (yych == 'o')
      goto yy466;
    goto yy364;
  yy429:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= 'D') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'E')
          goto yy467;
        if (yych == 'e')
          goto yy467;
        goto yy364;
      }
    }
  yy430:
    yych = *++p;
    if (yych == 'T')
      goto yy464;
    if (yych == 't')
      goto yy464;
    goto yy364;
  yy431:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= '@') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'A')
          goto yy468;
        if (yych == 'a')
          goto yy468;
        goto yy364;
      }
    }
  yy432:
    ++p;
    { return 2; }
  yy434:
    yych = *++p;
    if (yych == 'D')
      goto yy469;
    if (yych == 'd')
      goto yy469;
    goto yy364;
  yy435:
    yych = *++p;
    if (yych == 'R')
      goto yy470;
    if (yych == 'r')
      goto yy470;
    goto yy364;
  yy436:
    yych = *++p;
    if (yych == 'I')
      goto yy471;
    if (yych == 'i')
      goto yy471;
    goto yy364;
  yy437:
    yych = *++p;
    if (yych == 'D')
      goto yy472;
    if (yych == 'd')
      goto yy472;
    goto yy364;
  yy438:
    yych = *++p;
    if (yych == 'E')
      goto yy473;
    if (yych == 'e')
      goto yy473;
    goto yy364;
  yy439:
    yych = *++p;
    if (yych == 'C')
      goto yy474;
    if (yych == 'c')
      goto yy474;
    goto yy364;
  yy440:
    yych = *++p;
    if (yych == 'Y')
      goto yy399;
    if (yych == 'y')
      goto yy399;
    goto yy364;
  yy441:
    yych = *++p;
    if (yych == 'T')
      goto yy475;
    if (yych == 't')
      goto yy475;
    goto yy364;
  yy442:
    yych = *++p;
    if (yych == 'T')
      goto yy476;
    if (yych == 't')
      goto yy476;
    goto yy364;
  yy443:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= 'F') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'G')
          goto yy477;
        if (yych == 'g')
          goto yy477;
        goto yy364;
      }
    }
  yy444:
    yych = *++p;
    if (yych == 'A')
      goto yy478;
    if (yych == 'a')
      goto yy478;
    goto yy364;
  yy445:
    yych = *++p;
    if (yych == 'L')
      goto yy479;
    if (yych == 'l')
      goto yy479;
    goto yy364;
  yy446:
    yych = *++p;
    if (yych == 'L')
      goto yy480;
    if (yych == 'l')
      goto yy480;
    goto yy364;
  yy447:
    yych = *++p;
    if (yych <= 'U') {
      if (yych == 'C')
        goto yy481;
      if (yych <= 'T')
        goto yy364;
      goto yy482;
    } else {
      if (yych <= 'c') {
        if (yych <= 'b')
          goto yy364;
        goto yy481;
      } else {
        if (yych == 'u')
          goto yy482;
        goto yy364;
      }
    }
  yy448:
    yych = *++p;
    if (yych == 'M')
      goto yy399;
    if (yych == 'm')
      goto yy399;
    goto yy364;
  yy449:
    yych = *++p;
    if (yych == 'M')
      goto yy483;
    if (yych == 'm')
      goto yy483;
    goto yy364;
  yy450:
    yych = *++p;
    if (yych == 'D')
      goto yy484;
    if (yych == 'd')
      goto yy484;
    goto yy364;
  yy451:
    yych = *++p;
    if (yych == 'A')
      goto yy485;
    if (yych == 'a')
      goto yy485;
    goto yy364;
  yy452:
    yych = *++p;
    if (yych == 'E')
      goto yy486;
    if (yych == 'e')
      goto yy486;
    goto yy364;
  yy453:
    yych = *++p;
    if (yych == 'K')
      goto yy399;
    if (yych == 'k')
      goto yy399;
    goto yy364;
  yy454:
    yych = *++p;
    if (yych == 'N')
      goto yy399;
    if (yych == 'n')
      goto yy399;
    goto yy364;
  yy455:
    yych = *++p;
    if (yych == 'U')
      goto yy487;
    if (yych == 'u')
      goto yy487;
    goto yy364;
  yy456:
    yych = *++p;
    if (yych == 'R')
      goto yy488;
    if (yych == 'r')
      goto yy488;
    goto yy364;
  yy457:
    yych = *++p;
    if (yych <= 'I') {
      if (yych == 'G')
        goto yy477;
      if (yych <= 'H')
        goto yy364;
      goto yy489;
    } else {
      if (yych <= 'g') {
        if (yych <= 'f')
          goto yy364;
        goto yy477;
      } else {
        if (yych == 'i')
          goto yy489;
        goto yy364;
      }
    }
  yy458:
    yych = *++p;
    if (yych == 'A')
      goto yy448;
    if (yych == 'a')
      goto yy448;
    goto yy364;
  yy459:
    yych = *++p;
    if (yych <= 0x1F) {
      if (yych <= 0x08)
        goto yy364;
      if (yych <= '\r')
        goto yy490;
      goto yy364;
    } else {
      if (yych <= ' ')
        goto yy490;
      if (yych == '>')
        goto yy490;
      goto yy364;
    }
  yy460:
    yych = *++p;
    if (yych == 'I')
      goto yy492;
    if (yych == 'i')
      goto yy492;
    goto yy364;
  yy461:
    yych = *++p;
    if (yych == 'R')
      goto yy493;
    if (yych == 'r')
      goto yy493;
    goto yy364;
  yy462:
    yych = *++p;
    if (yych == 'L')
      goto yy419;
    if (yych == 'l')
      goto yy419;
    goto yy364;
  yy463:
    yych = *++p;
    if (yych == 'M')
      goto yy494;
    if (yych == 'm')
      goto yy494;
    goto yy364;
  yy464:
    yych = *++p;
    if (yych == 'L')
      goto yy472;
    if (yych == 'l')
      goto yy472;
    goto yy364;
  yy465:
    yych = *++p;
    if (yych == 'T')
      goto yy495;
    if (yych == 't')
      goto yy495;
    goto yy364;
  yy466:
    yych = *++p;
    if (yych == 'O')
      goto yy496;
    if (yych == 'o')
      goto yy496;
    goto yy364;
  yy467:
    yych = *++p;
    if (yych == 'A')
      goto yy497;
    if (yych == 'a')
      goto yy497;
    goto yy364;
  yy468:
    yych = *++p;
    if (yych == 'C')
      goto yy453;
    if (yych == 'c')
      goto yy453;
    goto yy364;
  yy469:
    yych = *++p;
    if (yych == 'A')
      goto yy498;
    if (yych == 'a')
      goto yy498;
    goto yy364;
  yy470:
    yych = *++p;
    if (yych == 'E')
      goto yy499;
    if (yych == 'e')
      goto yy499;
    goto yy364;
  yy471:
    yych = *++p;
    if (yych == 'C')
      goto yy464;
    if (yych == 'c')
      goto yy464;
    goto yy364;
  yy472:
    yych = *++p;
    if (yych == 'E')
      goto yy399;
    if (yych == 'e')
      goto yy399;
    goto yy364;
  yy473:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= 'E') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'F')
          goto yy500;
        if (yych == 'f')
          goto yy500;
        goto yy364;
      }
    }
  yy474:
    yych = *++p;
    if (yych == 'K')
      goto yy501;
    if (yych == 'k')
      goto yy501;
    goto yy364;
  yy475:
    yych = *++p;
    if (yych == 'I')
      goto yy489;
    if (yych == 'i')
      goto yy489;
    goto yy364;
  yy476:
    yych = *++p;
    if (yych == 'E')
      goto yy502;
    if (yych == 'e')
      goto yy502;
    goto yy364;
  yy477:
    yych = *++p;
    if (yych == 'R')
      goto yy503;
    if (yych == 'r')
      goto yy503;
    goto yy364;
  yy478:
    yych = *++p;
    if (yych == 'I')
      goto yy504;
    if (yych == 'i')
      goto yy504;
    goto yy364;
  yy479:
    yych = *++p;
    if (yych == 'O')
      goto yy505;
    if (yych == 'o')
      goto yy505;
    goto yy364;
  yy480:
    yych = *++p;
    if (yych == 'D')
      goto yy506;
    if (yych == 'd')
      goto yy506;
    goto yy364;
  yy481:
    yych = *++p;
    if (yych == 'A')
      goto yy396;
    if (yych == 'a')
      goto yy396;
    goto yy364;
  yy482:
    yych = *++p;
    if (yych == 'R')
      goto yy472;
    if (yych == 'r')
      goto yy472;
    goto yy364;
  yy483:
    yych = *++p;
    if (yych == 'E')
      goto yy507;
    if (yych == 'e')
      goto yy507;
    goto yy364;
  yy484:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= 'D') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'E')
          goto yy502;
        if (yych == 'e')
          goto yy502;
        goto yy364;
      }
    }
  yy485:
    yych = *++p;
    if (yych == 'M')
      goto yy472;
    if (yych == 'm')
      goto yy472;
    goto yy364;
  yy486:
    yych = *++p;
    if (yych == 'N')
      goto yy497;
    if (yych == 'n')
      goto yy497;
    goto yy364;
  yy487:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= 'H') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'I')
          goto yy508;
        if (yych == 'i')
          goto yy508;
        goto yy364;
      }
    }
  yy488:
    yych = *++p;
    if (yych == 'A')
      goto yy509;
    if (yych == 'a')
      goto yy509;
    goto yy364;
  yy489:
    yych = *++p;
    if (yych == 'O')
      goto yy454;
    if (yych == 'o')
      goto yy454;
    goto yy364;
  yy490:
    ++p;
    { return 1; }
  yy492:
    yych = *++p;
    if (yych == 'P')
      goto yy510;
    if (yych == 'p')
      goto yy510;
    goto yy364;
  yy493:
    yych = *++p;
    if (yych == 'C')
      goto yy472;
    if (yych == 'c')
      goto yy472;
    goto yy364;
  yy494:
    yych = *++p;
    if (yych == 'A')
      goto yy511;
    if (yych == 'a')
      goto yy511;
    goto yy364;
  yy495:
    yych = *++p;
    if (yych == 'A')
      goto yy512;
    if (yych == 'a')
      goto yy512;
    goto yy364;
  yy496:
    yych = *++p;
    if (yych == 'T')
      goto yy399;
    if (yych == 't')
      goto yy399;
    goto yy364;
  yy497:
    yych = *++p;
    if (yych == 'D')
      goto yy399;
    if (yych == 'd')
      goto yy399;
    goto yy364;
  yy498:
    yych = *++p;
    if (yych == 'T')
      goto yy513;
    if (yych == 't')
      goto yy513;
    goto yy364;
  yy499:
    yych = *++p;
    if (yych == 'S')
      goto yy514;
    if (yych == 's')
      goto yy514;
    goto yy364;
  yy500:
    yych = *++p;
    if (yych == 'O')
      goto yy515;
    if (yych == 'o')
      goto yy515;
    goto yy364;
  yy501:
    yych = *++p;
    if (yych == 'Q')
      goto yy516;
    if (yych == 'q')
      goto yy516;
    goto yy364;
  yy502:
    yych = *++p;
    if (yych == 'R')
      goto yy399;
    if (yych == 'r')
      goto yy399;
    goto yy364;
  yy503:
    yych = *++p;
    if (yych == 'O')
      goto yy517;
    if (yych == 'o')
      goto yy517;
    goto yy364;
  yy504:
    yych = *++p;
    if (yych == 'L')
      goto yy514;
    if (yych == 'l')
      goto yy514;
    goto yy364;
  yy505:
    yych = *++p;
    if (yych == 'G')
      goto yy399;
    if (yych == 'g')
      goto yy399;
    goto yy364;
  yy506:
    yych = *++p;
    if (yych == 'S')
      goto yy518;
    if (yych == 's')
      goto yy518;
    goto yy364;
  yy507:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy364;
        if (yych <= '\r')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= ' ')
          goto yy415;
        if (yych <= '.')
          goto yy364;
        goto yy417;
      }
    } else {
      if (yych <= 'R') {
        if (yych == '>')
          goto yy415;
        goto yy364;
      } else {
        if (yych <= 'S')
          goto yy518;
        if (yych == 's')
          goto yy518;
        goto yy364;
      }
    }
  yy508:
    yych = *++p;
    if (yych == 'T')
      goto yy519;
    if (yych == 't')
      goto yy519;
    goto yy364;
  yy509:
    yych = *++p;
    if (yych == 'M')
      goto yy520;
    if (yych == 'm')
      goto yy520;
    goto yy364;
  yy510:
    yych = *++p;
    if (yych == 'T')
      goto yy459;
    if (yych == 't')
      goto yy459;
    goto yy364;
  yy511:
    yych = *++p;
    if (yych == 'R')
      goto yy440;
    if (yych == 'r')
      goto yy440;
    goto yy364;
  yy512:
    yych = *++p;
    if (yych == 'R')
      goto yy521;
    if (yych == 'r')
      goto yy521;
    goto yy364;
  yy513:
    yych = *++p;
    if (yych == 'A')
      goto yy522;
    if (yych == 'a')
      goto yy522;
    goto yy364;
  yy514:
    yych = *++p;
    if (yych == 'S')
      goto yy399;
    if (yych == 's')
      goto yy399;
    goto yy364;
  yy515:
    yych = *++p;
    if (yych == 'N')
      goto yy496;
    if (yych == 'n')
      goto yy496;
    goto yy364;
  yy516:
    yych = *++p;
    if (yych == 'U')
      goto yy523;
    if (yych == 'u')
      goto yy523;
    goto yy364;
  yy517:
    yych = *++p;
    if (yych == 'U')
      goto yy524;
    if (yych == 'u')
      goto yy524;
    goto yy364;
  yy518:
    yych = *++p;
    if (yych == 'E')
      goto yy496;
    if (yych == 'e')
      goto yy496;
    goto yy364;
  yy519:
    yych = *++p;
    if (yych == 'E')
      goto yy448;
    if (yych == 'e')
      goto yy448;
    goto yy364;
  yy520:
    yych = *++p;
    if (yych == 'E')
      goto yy514;
    if (yych == 'e')
      goto yy514;
    goto yy364;
  yy521:
    yych = *++p;
    if (yych == 'E')
      goto yy525;
    if (yych == 'e')
      goto yy525;
    goto yy364;
  yy522:
    yych = *++p;
    if (yych == '[')
      goto yy526;
    goto yy364;
  yy523:
    yych = *++p;
    if (yych == 'O')
      goto yy528;
    if (yych == 'o')
      goto yy528;
    goto yy364;
  yy524:
    yych = *++p;
    if (yych == 'P')
      goto yy399;
    if (yych == 'p')
      goto yy399;
    goto yy364;
  yy525:
    yych = *++p;
    if (yych == 'A')
      goto yy459;
    if (yych == 'a')
      goto yy459;
    goto yy364;
  yy526:
    ++p;
    { return 5; }
  yy528:
    yych = *++p;
    if (yych == 'T')
      goto yy472;
    if (yych == 't')
      goto yy472;
    goto yy364;
  }
}

// Try to match an HTML block tag start line of type 7, returning
// 7 if successful, 0 if not.
bufsize_t _scan_html_block_start_7(const unsigned char *p) {
  const unsigned char *marker = NULL;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,   224, 224, 224, 224, 224, 224, 224, 224, 198, 210, 194, 198, 194,
        224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224, 224,
        224, 224, 224, 224, 198, 224, 128, 224, 224, 224, 224, 64,  224, 224,
        224, 224, 224, 233, 232, 224, 233, 233, 233, 233, 233, 233, 233, 233,
        233, 233, 232, 224, 192, 192, 192, 224, 224, 233, 233, 233, 233, 233,
        233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233,
        233, 233, 233, 233, 233, 233, 233, 224, 224, 224, 224, 232, 192, 233,
        233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233,
        233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 233, 224, 224, 224,
        224, 224, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych == '<')
      goto yy533;
    ++p;
  yy532 : { return 0; }
  yy533:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '@') {
      if (yych != '/')
        goto yy532;
    } else {
      if (yych <= 'Z')
        goto yy536;
      if (yych <= '`')
        goto yy532;
      if (yych <= 'z')
        goto yy536;
      goto yy532;
    }
    yych = *++p;
    if (yych <= '@')
      goto yy535;
    if (yych <= 'Z')
      goto yy538;
    if (yych <= '`')
      goto yy535;
    if (yych <= 'z')
      goto yy538;
  yy535:
    p = marker;
    if (yyaccept == 0) {
      goto yy532;
    } else {
      goto yy551;
    }
  yy536:
    yych = *++p;
    if (yybm[0 + yych] & 2) {
      goto yy540;
    }
    if (yych <= '=') {
      if (yych <= '.') {
        if (yych == '-')
          goto yy536;
        goto yy535;
      } else {
        if (yych <= '/')
          goto yy542;
        if (yych <= '9')
          goto yy536;
        goto yy535;
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '>')
          goto yy543;
        if (yych <= '@')
          goto yy535;
        goto yy536;
      } else {
        if (yych <= '`')
          goto yy535;
        if (yych <= 'z')
          goto yy536;
        goto yy535;
      }
    }
  yy538:
    yych = *++p;
    if (yych <= '/') {
      if (yych <= 0x1F) {
        if (yych <= 0x08)
          goto yy535;
        if (yych <= '\r')
          goto yy545;
        goto yy535;
      } else {
        if (yych <= ' ')
          goto yy545;
        if (yych == '-')
          goto yy538;
        goto yy535;
      }
    } else {
      if (yych <= '@') {
        if (yych <= '9')
          goto yy538;
        if (yych == '>')
          goto yy543;
        goto yy535;
      } else {
        if (yych <= 'Z')
          goto yy538;
        if (yych <= '`')
          goto yy535;
        if (yych <= 'z')
          goto yy538;
        goto yy535;
      }
    }
  yy540:
    yych = *++p;
    if (yybm[0 + yych] & 2) {
      goto yy540;
    }
    if (yych <= '>') {
      if (yych <= '9') {
        if (yych != '/')
          goto yy535;
      } else {
        if (yych <= ':')
          goto yy547;
        if (yych <= '=')
          goto yy535;
        goto yy543;
      }
    } else {
      if (yych <= '^') {
        if (yych <= '@')
          goto yy535;
        if (yych <= 'Z')
          goto yy547;
        goto yy535;
      } else {
        if (yych == '`')
          goto yy535;
        if (yych <= 'z')
          goto yy547;
        goto yy535;
      }
    }
  yy542:
    yych = *++p;
    if (yych != '>')
      goto yy535;
  yy543:
    yych = *++p;
    if (yybm[0 + yych] & 4) {
      goto yy543;
    }
    if (yych <= 0x08)
      goto yy535;
    if (yych <= '\n')
      goto yy549;
    if (yych <= '\v')
      goto yy535;
    if (yych <= '\r')
      goto yy552;
    goto yy535;
  yy545:
    yych = *++p;
    if (yych <= 0x1F) {
      if (yych <= 0x08)
        goto yy535;
      if (yych <= '\r')
        goto yy545;
      goto yy535;
    } else {
      if (yych <= ' ')
        goto yy545;
      if (yych == '>')
        goto yy543;
      goto yy535;
    }
  yy547:
    yych = *++p;
    if (yybm[0 + yych] & 8) {
      goto yy547;
    }
    if (yych <= ',') {
      if (yych <= '\r') {
        if (yych <= 0x08)
          goto yy535;
        goto yy553;
      } else {
        if (yych == ' ')
          goto yy553;
        goto yy535;
      }
    } else {
      if (yych <= '<') {
        if (yych <= '/')
          goto yy542;
        goto yy535;
      } else {
        if (yych <= '=')
          goto yy555;
        if (yych <= '>')
          goto yy543;
        goto yy535;
      }
    }
  yy549:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 4) {
      goto yy543;
    }
    if (yych <= 0x08)
      goto yy551;
    if (yych <= '\n')
      goto yy549;
    if (yych <= '\v')
      goto yy551;
    if (yych <= '\r')
      goto yy552;
  yy551 : { return 7; }
  yy552:
    ++p;
    goto yy551;
  yy553:
    yych = *++p;
    if (yych <= '<') {
      if (yych <= ' ') {
        if (yych <= 0x08)
          goto yy535;
        if (yych <= '\r')
          goto yy553;
        if (yych <= 0x1F)
          goto yy535;
        goto yy553;
      } else {
        if (yych <= '/') {
          if (yych <= '.')
            goto yy535;
          goto yy542;
        } else {
          if (yych == ':')
            goto yy547;
          goto yy535;
        }
      }
    } else {
      if (yych <= 'Z') {
        if (yych <= '=')
          goto yy555;
        if (yych <= '>')
          goto yy543;
        if (yych <= '@')
          goto yy535;
        goto yy547;
      } else {
        if (yych <= '_') {
          if (yych <= '^')
            goto yy535;
          goto yy547;
        } else {
          if (yych <= '`')
            goto yy535;
          if (yych <= 'z')
            goto yy547;
          goto yy535;
        }
      }
    }
  yy555:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy557;
    }
    if (yych <= 0xE0) {
      if (yych <= '"') {
        if (yych <= 0x00)
          goto yy535;
        if (yych <= ' ')
          goto yy555;
        goto yy559;
      } else {
        if (yych <= '\'')
          goto yy561;
        if (yych <= 0xC1)
          goto yy535;
        if (yych <= 0xDF)
          goto yy563;
        goto yy564;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy566;
        goto yy565;
      } else {
        if (yych <= 0xF0)
          goto yy567;
        if (yych <= 0xF3)
          goto yy568;
        if (yych <= 0xF4)
          goto yy569;
        goto yy535;
      }
    }
  yy557:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy557;
    }
    if (yych <= 0xE0) {
      if (yych <= '=') {
        if (yych <= 0x00)
          goto yy535;
        if (yych <= ' ')
          goto yy540;
        goto yy535;
      } else {
        if (yych <= '>')
          goto yy543;
        if (yych <= 0xC1)
          goto yy535;
        if (yych <= 0xDF)
          goto yy563;
        goto yy564;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy566;
        goto yy565;
      } else {
        if (yych <= 0xF0)
          goto yy567;
        if (yych <= 0xF3)
          goto yy568;
        if (yych <= 0xF4)
          goto yy569;
        goto yy535;
      }
    }
  yy559:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy559;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy535;
        if (yych <= '"')
          goto yy570;
        goto yy535;
      } else {
        if (yych <= 0xDF)
          goto yy571;
        if (yych <= 0xE0)
          goto yy572;
        goto yy573;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy574;
        if (yych <= 0xEF)
          goto yy573;
        goto yy575;
      } else {
        if (yych <= 0xF3)
          goto yy576;
        if (yych <= 0xF4)
          goto yy577;
        goto yy535;
      }
    }
  yy561:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy561;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy535;
        if (yych <= '\'')
          goto yy570;
        goto yy535;
      } else {
        if (yych <= 0xDF)
          goto yy578;
        if (yych <= 0xE0)
          goto yy579;
        goto yy580;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy581;
        if (yych <= 0xEF)
          goto yy580;
        goto yy582;
      } else {
        if (yych <= 0xF3)
          goto yy583;
        if (yych <= 0xF4)
          goto yy584;
        goto yy535;
      }
    }
  yy563:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy557;
    goto yy535;
  yy564:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy563;
    goto yy535;
  yy565:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy563;
    goto yy535;
  yy566:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0x9F)
      goto yy563;
    goto yy535;
  yy567:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy565;
    goto yy535;
  yy568:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy565;
    goto yy535;
  yy569:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0x8F)
      goto yy565;
    goto yy535;
  yy570:
    yych = *++p;
    if (yybm[0 + yych] & 2) {
      goto yy540;
    }
    if (yych == '/')
      goto yy542;
    if (yych == '>')
      goto yy543;
    goto yy535;
  yy571:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy559;
    goto yy535;
  yy572:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy571;
    goto yy535;
  yy573:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy571;
    goto yy535;
  yy574:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0x9F)
      goto yy571;
    goto yy535;
  yy575:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy573;
    goto yy535;
  yy576:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy573;
    goto yy535;
  yy577:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0x8F)
      goto yy573;
    goto yy535;
  yy578:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy561;
    goto yy535;
  yy579:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy578;
    goto yy535;
  yy580:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy578;
    goto yy535;
  yy581:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0x9F)
      goto yy578;
    goto yy535;
  yy582:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy580;
    goto yy535;
  yy583:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0xBF)
      goto yy580;
    goto yy535;
  yy584:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy535;
    if (yych <= 0x8F)
      goto yy580;
    goto yy535;
  }
}

// Try to match an HTML block end line of type 1
bufsize_t _scan_html_block_end_1(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,  64, 64, 64, 64, 64, 64,  64, 64, 64, 0,  64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 128, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,
    };
    yych = *p;
    if (yych <= 0xDF) {
      if (yych <= ';') {
        if (yych <= 0x00)
          goto yy587;
        if (yych != '\n')
          goto yy589;
      } else {
        if (yych <= '<')
          goto yy590;
        if (yych <= 0x7F)
          goto yy589;
        if (yych >= 0xC2)
          goto yy591;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy592;
        if (yych == 0xED)
          goto yy594;
        goto yy593;
      } else {
        if (yych <= 0xF0)
          goto yy595;
        if (yych <= 0xF3)
          goto yy596;
        if (yych <= 0xF4)
          goto yy597;
      }
    }
  yy587:
    ++p;
  yy588 : { return 0; }
  yy589:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '\n') {
      if (yych <= 0x00)
        goto yy588;
      if (yych <= '\t')
        goto yy599;
      goto yy588;
    } else {
      if (yych <= 0x7F)
        goto yy599;
      if (yych <= 0xC1)
        goto yy588;
      if (yych <= 0xF4)
        goto yy599;
      goto yy588;
    }
  yy590:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '.') {
      if (yych <= 0x00)
        goto yy588;
      if (yych == '\n')
        goto yy588;
      goto yy599;
    } else {
      if (yych <= 0x7F) {
        if (yych <= '/')
          goto yy610;
        goto yy599;
      } else {
        if (yych <= 0xC1)
          goto yy588;
        if (yych <= 0xF4)
          goto yy599;
        goto yy588;
      }
    }
  yy591:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy588;
    if (yych <= 0xBF)
      goto yy598;
    goto yy588;
  yy592:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x9F)
      goto yy588;
    if (yych <= 0xBF)
      goto yy603;
    goto yy588;
  yy593:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy588;
    if (yych <= 0xBF)
      goto yy603;
    goto yy588;
  yy594:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy588;
    if (yych <= 0x9F)
      goto yy603;
    goto yy588;
  yy595:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x8F)
      goto yy588;
    if (yych <= 0xBF)
      goto yy605;
    goto yy588;
  yy596:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy588;
    if (yych <= 0xBF)
      goto yy605;
    goto yy588;
  yy597:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy588;
    if (yych <= 0x8F)
      goto yy605;
    goto yy588;
  yy598:
    yych = *++p;
  yy599:
    if (yybm[0 + yych] & 64) {
      goto yy598;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy600;
        if (yych <= '<')
          goto yy601;
      } else {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        goto yy605;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy606;
        if (yych <= 0xEF)
          goto yy605;
        goto yy607;
      } else {
        if (yych <= 0xF3)
          goto yy608;
        if (yych <= 0xF4)
          goto yy609;
      }
    }
  yy600:
    p = marker;
    if (yyaccept == 0) {
      goto yy588;
    } else {
      goto yy623;
    }
  yy601:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xDF) {
      if (yych <= '.') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= '/')
          goto yy610;
        if (yych <= 0x7F)
          goto yy598;
        if (yych <= 0xC1)
          goto yy600;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy604;
        if (yych == 0xED)
          goto yy606;
        goto yy605;
      } else {
        if (yych <= 0xF0)
          goto yy607;
        if (yych <= 0xF3)
          goto yy608;
        if (yych <= 0xF4)
          goto yy609;
        goto yy600;
      }
    }
  yy603:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy600;
    if (yych <= 0xBF)
      goto yy598;
    goto yy600;
  yy604:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy600;
    if (yych <= 0xBF)
      goto yy603;
    goto yy600;
  yy605:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy600;
    if (yych <= 0xBF)
      goto yy603;
    goto yy600;
  yy606:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy600;
    if (yych <= 0x9F)
      goto yy603;
    goto yy600;
  yy607:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy600;
    if (yych <= 0xBF)
      goto yy605;
    goto yy600;
  yy608:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy600;
    if (yych <= 0xBF)
      goto yy605;
    goto yy600;
  yy609:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy600;
    if (yych <= 0x8F)
      goto yy605;
    goto yy600;
  yy610:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 's') {
      if (yych <= 'R') {
        if (yych <= '\n') {
          if (yych <= 0x00)
            goto yy600;
          if (yych <= '\t')
            goto yy598;
          goto yy600;
        } else {
          if (yych != 'P')
            goto yy598;
        }
      } else {
        if (yych <= 'o') {
          if (yych <= 'S')
            goto yy612;
          if (yych <= 'T')
            goto yy613;
          goto yy598;
        } else {
          if (yych <= 'p')
            goto yy611;
          if (yych <= 'r')
            goto yy598;
          goto yy612;
        }
      }
    } else {
      if (yych <= 0xEC) {
        if (yych <= 0xC1) {
          if (yych <= 't')
            goto yy613;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        } else {
          if (yych <= 0xDF)
            goto yy603;
          if (yych <= 0xE0)
            goto yy604;
          goto yy605;
        }
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xED)
            goto yy606;
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy611:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'Q') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'q') {
          if (yych <= 'R')
            goto yy614;
          goto yy598;
        } else {
          if (yych <= 'r')
            goto yy614;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy612:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 't') {
      if (yych <= 'C') {
        if (yych <= '\t') {
          if (yych <= 0x00)
            goto yy600;
          goto yy598;
        } else {
          if (yych <= '\n')
            goto yy600;
          if (yych <= 'B')
            goto yy598;
          goto yy615;
        }
      } else {
        if (yych <= 'b') {
          if (yych == 'T')
            goto yy616;
          goto yy598;
        } else {
          if (yych <= 'c')
            goto yy615;
          if (yych <= 's')
            goto yy598;
          goto yy616;
        }
      }
    } else {
      if (yych <= 0xEC) {
        if (yych <= 0xC1) {
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        } else {
          if (yych <= 0xDF)
            goto yy603;
          if (yych <= 0xE0)
            goto yy604;
          goto yy605;
        }
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xED)
            goto yy606;
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy613:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'D') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'd') {
          if (yych <= 'E')
            goto yy617;
          goto yy598;
        } else {
          if (yych <= 'e')
            goto yy617;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy614:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'D') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'd') {
          if (yych <= 'E')
            goto yy618;
          goto yy598;
        } else {
          if (yych <= 'e')
            goto yy618;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy615:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'Q') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'q') {
          if (yych <= 'R')
            goto yy619;
          goto yy598;
        } else {
          if (yych <= 'r')
            goto yy619;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy616:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'X') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'x') {
          if (yych <= 'Y')
            goto yy620;
          goto yy598;
        } else {
          if (yych <= 'y')
            goto yy620;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy617:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'W') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'w') {
          if (yych <= 'X')
            goto yy621;
          goto yy598;
        } else {
          if (yych <= 'x')
            goto yy621;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy618:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xDF) {
      if (yych <= '=') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= '>')
          goto yy622;
        if (yych <= 0x7F)
          goto yy598;
        if (yych <= 0xC1)
          goto yy600;
        goto yy603;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy604;
        if (yych == 0xED)
          goto yy606;
        goto yy605;
      } else {
        if (yych <= 0xF0)
          goto yy607;
        if (yych <= 0xF3)
          goto yy608;
        if (yych <= 0xF4)
          goto yy609;
        goto yy600;
      }
    }
  yy619:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'H') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'h') {
          if (yych <= 'I')
            goto yy624;
          goto yy598;
        } else {
          if (yych <= 'i')
            goto yy624;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy620:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'K') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'k') {
          if (yych <= 'L')
            goto yy614;
          goto yy598;
        } else {
          if (yych <= 'l')
            goto yy614;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy621:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'S') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 's') {
          if (yych <= 'T')
            goto yy625;
          goto yy598;
        } else {
          if (yych <= 't')
            goto yy625;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy622:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 64) {
      goto yy598;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy623;
        if (yych <= '<')
          goto yy601;
      } else {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        goto yy605;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy606;
        if (yych <= 0xEF)
          goto yy605;
        goto yy607;
      } else {
        if (yych <= 0xF3)
          goto yy608;
        if (yych <= 0xF4)
          goto yy609;
      }
    }
  yy623 : { return (bufsize_t)(p - start); }
  yy624:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'O') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'o') {
          if (yych <= 'P')
            goto yy626;
          goto yy598;
        } else {
          if (yych <= 'p')
            goto yy626;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy625:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= '@') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= '`') {
          if (yych <= 'A')
            goto yy627;
          goto yy598;
        } else {
          if (yych <= 'a')
            goto yy627;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy626:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'S') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 's') {
          if (yych <= 'T')
            goto yy618;
          goto yy598;
        } else {
          if (yych <= 't')
            goto yy618;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy627:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'Q') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'q') {
          if (yych >= 'S')
            goto yy598;
        } else {
          if (yych <= 'r')
            goto yy628;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy628:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= 'D') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= 'd') {
          if (yych >= 'F')
            goto yy598;
        } else {
          if (yych <= 'e')
            goto yy629;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  yy629:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy601;
    }
    if (yych <= 0xC1) {
      if (yych <= '@') {
        if (yych <= 0x00)
          goto yy600;
        if (yych == '\n')
          goto yy600;
        goto yy598;
      } else {
        if (yych <= '`') {
          if (yych <= 'A')
            goto yy618;
          goto yy598;
        } else {
          if (yych <= 'a')
            goto yy618;
          if (yych <= 0x7F)
            goto yy598;
          goto yy600;
        }
      }
    } else {
      if (yych <= 0xED) {
        if (yych <= 0xDF)
          goto yy603;
        if (yych <= 0xE0)
          goto yy604;
        if (yych <= 0xEC)
          goto yy605;
        goto yy606;
      } else {
        if (yych <= 0xF0) {
          if (yych <= 0xEF)
            goto yy605;
          goto yy607;
        } else {
          if (yych <= 0xF3)
            goto yy608;
          if (yych <= 0xF4)
            goto yy609;
          goto yy600;
        }
      }
    }
  }
}

// Try to match an HTML block end line of type 2
bufsize_t _scan_html_block_end_2(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,  64, 64, 64, 64, 64, 64, 64, 64, 64,  0,  64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 128, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,
    };
    yych = *p;
    if (yych <= 0xDF) {
      if (yych <= ',') {
        if (yych <= 0x00)
          goto yy632;
        if (yych != '\n')
          goto yy634;
      } else {
        if (yych <= '-')
          goto yy635;
        if (yych <= 0x7F)
          goto yy634;
        if (yych >= 0xC2)
          goto yy636;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy637;
        if (yych == 0xED)
          goto yy639;
        goto yy638;
      } else {
        if (yych <= 0xF0)
          goto yy640;
        if (yych <= 0xF3)
          goto yy641;
        if (yych <= 0xF4)
          goto yy642;
      }
    }
  yy632:
    ++p;
  yy633 : { return 0; }
  yy634:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '\n') {
      if (yych <= 0x00)
        goto yy633;
      if (yych <= '\t')
        goto yy644;
      goto yy633;
    } else {
      if (yych <= 0x7F)
        goto yy644;
      if (yych <= 0xC1)
        goto yy633;
      if (yych <= 0xF4)
        goto yy644;
      goto yy633;
    }
  yy635:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy654;
    }
    if (yych <= '\n') {
      if (yych <= 0x00)
        goto yy633;
      if (yych <= '\t')
        goto yy644;
      goto yy633;
    } else {
      if (yych <= 0x7F)
        goto yy644;
      if (yych <= 0xC1)
        goto yy633;
      if (yych <= 0xF4)
        goto yy644;
      goto yy633;
    }
  yy636:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy633;
    if (yych <= 0xBF)
      goto yy643;
    goto yy633;
  yy637:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x9F)
      goto yy633;
    if (yych <= 0xBF)
      goto yy647;
    goto yy633;
  yy638:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy633;
    if (yych <= 0xBF)
      goto yy647;
    goto yy633;
  yy639:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy633;
    if (yych <= 0x9F)
      goto yy647;
    goto yy633;
  yy640:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x8F)
      goto yy633;
    if (yych <= 0xBF)
      goto yy649;
    goto yy633;
  yy641:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy633;
    if (yych <= 0xBF)
      goto yy649;
    goto yy633;
  yy642:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy633;
    if (yych <= 0x8F)
      goto yy649;
    goto yy633;
  yy643:
    yych = *++p;
  yy644:
    if (yybm[0 + yych] & 64) {
      goto yy643;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy645;
        if (yych <= '-')
          goto yy646;
      } else {
        if (yych <= 0xDF)
          goto yy647;
        if (yych <= 0xE0)
          goto yy648;
        goto yy649;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy650;
        if (yych <= 0xEF)
          goto yy649;
        goto yy651;
      } else {
        if (yych <= 0xF3)
          goto yy652;
        if (yych <= 0xF4)
          goto yy653;
      }
    }
  yy645:
    p = marker;
    if (yyaccept == 0) {
      goto yy633;
    } else {
      goto yy657;
    }
  yy646:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy643;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy645;
        if (yych <= '-')
          goto yy654;
        goto yy645;
      } else {
        if (yych <= 0xDF)
          goto yy647;
        if (yych <= 0xE0)
          goto yy648;
        goto yy649;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy650;
        if (yych <= 0xEF)
          goto yy649;
        goto yy651;
      } else {
        if (yych <= 0xF3)
          goto yy652;
        if (yych <= 0xF4)
          goto yy653;
        goto yy645;
      }
    }
  yy647:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy645;
    if (yych <= 0xBF)
      goto yy643;
    goto yy645;
  yy648:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy645;
    if (yych <= 0xBF)
      goto yy647;
    goto yy645;
  yy649:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy645;
    if (yych <= 0xBF)
      goto yy647;
    goto yy645;
  yy650:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy645;
    if (yych <= 0x9F)
      goto yy647;
    goto yy645;
  yy651:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy645;
    if (yych <= 0xBF)
      goto yy649;
    goto yy645;
  yy652:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy645;
    if (yych <= 0xBF)
      goto yy649;
    goto yy645;
  yy653:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy645;
    if (yych <= 0x8F)
      goto yy649;
    goto yy645;
  yy654:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy654;
    }
    if (yych <= 0xDF) {
      if (yych <= '=') {
        if (yych <= 0x00)
          goto yy645;
        if (yych == '\n')
          goto yy645;
        goto yy643;
      } else {
        if (yych <= '>')
          goto yy656;
        if (yych <= 0x7F)
          goto yy643;
        if (yych <= 0xC1)
          goto yy645;
        goto yy647;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy648;
        if (yych == 0xED)
          goto yy650;
        goto yy649;
      } else {
        if (yych <= 0xF0)
          goto yy651;
        if (yych <= 0xF3)
          goto yy652;
        if (yych <= 0xF4)
          goto yy653;
        goto yy645;
      }
    }
  yy656:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 64) {
      goto yy643;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy657;
        if (yych <= '-')
          goto yy646;
      } else {
        if (yych <= 0xDF)
          goto yy647;
        if (yych <= 0xE0)
          goto yy648;
        goto yy649;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy650;
        if (yych <= 0xEF)
          goto yy649;
        goto yy651;
      } else {
        if (yych <= 0xF3)
          goto yy652;
        if (yych <= 0xF4)
          goto yy653;
      }
    }
  yy657 : { return (bufsize_t)(p - start); }
  }
}

// Try to match an HTML block end line of type 3
bufsize_t _scan_html_block_end_3(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,  64, 64, 64, 64, 64, 64, 64, 64, 64,  0,  64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 128, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64, 64, 64, 64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,
    };
    yych = *p;
    if (yych <= 0xDF) {
      if (yych <= '>') {
        if (yych <= 0x00)
          goto yy660;
        if (yych != '\n')
          goto yy662;
      } else {
        if (yych <= '?')
          goto yy663;
        if (yych <= 0x7F)
          goto yy662;
        if (yych >= 0xC2)
          goto yy664;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy665;
        if (yych == 0xED)
          goto yy667;
        goto yy666;
      } else {
        if (yych <= 0xF0)
          goto yy668;
        if (yych <= 0xF3)
          goto yy669;
        if (yych <= 0xF4)
          goto yy670;
      }
    }
  yy660:
    ++p;
  yy661 : { return 0; }
  yy662:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '\n') {
      if (yych <= 0x00)
        goto yy661;
      if (yych <= '\t')
        goto yy672;
      goto yy661;
    } else {
      if (yych <= 0x7F)
        goto yy672;
      if (yych <= 0xC1)
        goto yy661;
      if (yych <= 0xF4)
        goto yy672;
      goto yy661;
    }
  yy663:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '=') {
      if (yych <= 0x00)
        goto yy661;
      if (yych == '\n')
        goto yy661;
      goto yy672;
    } else {
      if (yych <= 0x7F) {
        if (yych <= '>')
          goto yy683;
        goto yy672;
      } else {
        if (yych <= 0xC1)
          goto yy661;
        if (yych <= 0xF4)
          goto yy672;
        goto yy661;
      }
    }
  yy664:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy661;
    if (yych <= 0xBF)
      goto yy671;
    goto yy661;
  yy665:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x9F)
      goto yy661;
    if (yych <= 0xBF)
      goto yy676;
    goto yy661;
  yy666:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy661;
    if (yych <= 0xBF)
      goto yy676;
    goto yy661;
  yy667:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy661;
    if (yych <= 0x9F)
      goto yy676;
    goto yy661;
  yy668:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x8F)
      goto yy661;
    if (yych <= 0xBF)
      goto yy678;
    goto yy661;
  yy669:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy661;
    if (yych <= 0xBF)
      goto yy678;
    goto yy661;
  yy670:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy661;
    if (yych <= 0x8F)
      goto yy678;
    goto yy661;
  yy671:
    yych = *++p;
  yy672:
    if (yybm[0 + yych] & 64) {
      goto yy671;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy673;
        if (yych <= '?')
          goto yy674;
      } else {
        if (yych <= 0xDF)
          goto yy676;
        if (yych <= 0xE0)
          goto yy677;
        goto yy678;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy679;
        if (yych <= 0xEF)
          goto yy678;
        goto yy680;
      } else {
        if (yych <= 0xF3)
          goto yy681;
        if (yych <= 0xF4)
          goto yy682;
      }
    }
  yy673:
    p = marker;
    if (yyaccept == 0) {
      goto yy661;
    } else {
      goto yy684;
    }
  yy674:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy674;
    }
    if (yych <= 0xDF) {
      if (yych <= '=') {
        if (yych <= 0x00)
          goto yy673;
        if (yych == '\n')
          goto yy673;
        goto yy671;
      } else {
        if (yych <= '>')
          goto yy683;
        if (yych <= 0x7F)
          goto yy671;
        if (yych <= 0xC1)
          goto yy673;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy677;
        if (yych == 0xED)
          goto yy679;
        goto yy678;
      } else {
        if (yych <= 0xF0)
          goto yy680;
        if (yych <= 0xF3)
          goto yy681;
        if (yych <= 0xF4)
          goto yy682;
        goto yy673;
      }
    }
  yy676:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy673;
    if (yych <= 0xBF)
      goto yy671;
    goto yy673;
  yy677:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy673;
    if (yych <= 0xBF)
      goto yy676;
    goto yy673;
  yy678:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy673;
    if (yych <= 0xBF)
      goto yy676;
    goto yy673;
  yy679:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy673;
    if (yych <= 0x9F)
      goto yy676;
    goto yy673;
  yy680:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy673;
    if (yych <= 0xBF)
      goto yy678;
    goto yy673;
  yy681:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy673;
    if (yych <= 0xBF)
      goto yy678;
    goto yy673;
  yy682:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy673;
    if (yych <= 0x8F)
      goto yy678;
    goto yy673;
  yy683:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 64) {
      goto yy671;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy684;
        if (yych <= '?')
          goto yy674;
      } else {
        if (yych <= 0xDF)
          goto yy676;
        if (yych <= 0xE0)
          goto yy677;
        goto yy678;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy679;
        if (yych <= 0xEF)
          goto yy678;
        goto yy680;
      } else {
        if (yych <= 0xF3)
          goto yy681;
        if (yych <= 0xF4)
          goto yy682;
      }
    }
  yy684 : { return (bufsize_t)(p - start); }
  }
}

// Try to match an HTML block end line of type 4
bufsize_t _scan_html_block_end_4(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,   128, 128, 128, 128, 128, 128, 128, 128, 128, 0,   128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 64,  128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128, 128,
        128, 128, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yybm[0 + yych] & 64) {
      goto yy690;
    }
    if (yych <= 0xE0) {
      if (yych <= '\n') {
        if (yych <= 0x00)
          goto yy687;
        if (yych <= '\t')
          goto yy689;
      } else {
        if (yych <= 0x7F)
          goto yy689;
        if (yych <= 0xC1)
          goto yy687;
        if (yych <= 0xDF)
          goto yy693;
        goto yy694;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy696;
        goto yy695;
      } else {
        if (yych <= 0xF0)
          goto yy697;
        if (yych <= 0xF3)
          goto yy698;
        if (yych <= 0xF4)
          goto yy699;
      }
    }
  yy687:
    ++p;
  yy688 : { return 0; }
  yy689:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '\n') {
      if (yych <= 0x00)
        goto yy688;
      if (yych <= '\t')
        goto yy701;
      goto yy688;
    } else {
      if (yych <= 0x7F)
        goto yy701;
      if (yych <= 0xC1)
        goto yy688;
      if (yych <= 0xF4)
        goto yy701;
      goto yy688;
    }
  yy690:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy700;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy692;
        if (yych <= '>')
          goto yy690;
      } else {
        if (yych <= 0xDF)
          goto yy703;
        if (yych <= 0xE0)
          goto yy704;
        goto yy705;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy706;
        if (yych <= 0xEF)
          goto yy705;
        goto yy707;
      } else {
        if (yych <= 0xF3)
          goto yy708;
        if (yych <= 0xF4)
          goto yy709;
      }
    }
  yy692 : { return (bufsize_t)(p - start); }
  yy693:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy688;
    if (yych <= 0xBF)
      goto yy700;
    goto yy688;
  yy694:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x9F)
      goto yy688;
    if (yych <= 0xBF)
      goto yy703;
    goto yy688;
  yy695:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy688;
    if (yych <= 0xBF)
      goto yy703;
    goto yy688;
  yy696:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy688;
    if (yych <= 0x9F)
      goto yy703;
    goto yy688;
  yy697:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x8F)
      goto yy688;
    if (yych <= 0xBF)
      goto yy705;
    goto yy688;
  yy698:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy688;
    if (yych <= 0xBF)
      goto yy705;
    goto yy688;
  yy699:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy688;
    if (yych <= 0x8F)
      goto yy705;
    goto yy688;
  yy700:
    yych = *++p;
  yy701:
    if (yybm[0 + yych] & 128) {
      goto yy700;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy702;
        if (yych <= '>')
          goto yy690;
      } else {
        if (yych <= 0xDF)
          goto yy703;
        if (yych <= 0xE0)
          goto yy704;
        goto yy705;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy706;
        if (yych <= 0xEF)
          goto yy705;
        goto yy707;
      } else {
        if (yych <= 0xF3)
          goto yy708;
        if (yych <= 0xF4)
          goto yy709;
      }
    }
  yy702:
    p = marker;
    if (yyaccept == 0) {
      goto yy688;
    } else {
      goto yy692;
    }
  yy703:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy702;
    if (yych <= 0xBF)
      goto yy700;
    goto yy702;
  yy704:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy702;
    if (yych <= 0xBF)
      goto yy703;
    goto yy702;
  yy705:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy702;
    if (yych <= 0xBF)
      goto yy703;
    goto yy702;
  yy706:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy702;
    if (yych <= 0x9F)
      goto yy703;
    goto yy702;
  yy707:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy702;
    if (yych <= 0xBF)
      goto yy705;
    goto yy702;
  yy708:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy702;
    if (yych <= 0xBF)
      goto yy705;
    goto yy702;
  yy709:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy702;
    if (yych <= 0x8F)
      goto yy705;
    goto yy702;
  }
}

// Try to match an HTML block end line of type 5
bufsize_t _scan_html_block_end_5(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,  64, 64, 64,  64, 64, 64, 64, 64, 64, 0,  64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 128, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 64, 64,  64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64, 64,
        64, 64, 0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,
    };
    yych = *p;
    if (yych <= 0xDF) {
      if (yych <= '\\') {
        if (yych <= 0x00)
          goto yy712;
        if (yych != '\n')
          goto yy714;
      } else {
        if (yych <= ']')
          goto yy715;
        if (yych <= 0x7F)
          goto yy714;
        if (yych >= 0xC2)
          goto yy716;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy717;
        if (yych == 0xED)
          goto yy719;
        goto yy718;
      } else {
        if (yych <= 0xF0)
          goto yy720;
        if (yych <= 0xF3)
          goto yy721;
        if (yych <= 0xF4)
          goto yy722;
      }
    }
  yy712:
    ++p;
  yy713 : { return 0; }
  yy714:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '\n') {
      if (yych <= 0x00)
        goto yy713;
      if (yych <= '\t')
        goto yy724;
      goto yy713;
    } else {
      if (yych <= 0x7F)
        goto yy724;
      if (yych <= 0xC1)
        goto yy713;
      if (yych <= 0xF4)
        goto yy724;
      goto yy713;
    }
  yy715:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy734;
    }
    if (yych <= '\n') {
      if (yych <= 0x00)
        goto yy713;
      if (yych <= '\t')
        goto yy724;
      goto yy713;
    } else {
      if (yych <= 0x7F)
        goto yy724;
      if (yych <= 0xC1)
        goto yy713;
      if (yych <= 0xF4)
        goto yy724;
      goto yy713;
    }
  yy716:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy713;
    if (yych <= 0xBF)
      goto yy723;
    goto yy713;
  yy717:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x9F)
      goto yy713;
    if (yych <= 0xBF)
      goto yy727;
    goto yy713;
  yy718:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy713;
    if (yych <= 0xBF)
      goto yy727;
    goto yy713;
  yy719:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy713;
    if (yych <= 0x9F)
      goto yy727;
    goto yy713;
  yy720:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x8F)
      goto yy713;
    if (yych <= 0xBF)
      goto yy729;
    goto yy713;
  yy721:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy713;
    if (yych <= 0xBF)
      goto yy729;
    goto yy713;
  yy722:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x7F)
      goto yy713;
    if (yych <= 0x8F)
      goto yy729;
    goto yy713;
  yy723:
    yych = *++p;
  yy724:
    if (yybm[0 + yych] & 64) {
      goto yy723;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy725;
        if (yych <= ']')
          goto yy726;
      } else {
        if (yych <= 0xDF)
          goto yy727;
        if (yych <= 0xE0)
          goto yy728;
        goto yy729;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy730;
        if (yych <= 0xEF)
          goto yy729;
        goto yy731;
      } else {
        if (yych <= 0xF3)
          goto yy732;
        if (yych <= 0xF4)
          goto yy733;
      }
    }
  yy725:
    p = marker;
    if (yyaccept == 0) {
      goto yy713;
    } else {
      goto yy737;
    }
  yy726:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy723;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy725;
        if (yych <= ']')
          goto yy734;
        goto yy725;
      } else {
        if (yych <= 0xDF)
          goto yy727;
        if (yych <= 0xE0)
          goto yy728;
        goto yy729;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy730;
        if (yych <= 0xEF)
          goto yy729;
        goto yy731;
      } else {
        if (yych <= 0xF3)
          goto yy732;
        if (yych <= 0xF4)
          goto yy733;
        goto yy725;
      }
    }
  yy727:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy725;
    if (yych <= 0xBF)
      goto yy723;
    goto yy725;
  yy728:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy725;
    if (yych <= 0xBF)
      goto yy727;
    goto yy725;
  yy729:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy725;
    if (yych <= 0xBF)
      goto yy727;
    goto yy725;
  yy730:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy725;
    if (yych <= 0x9F)
      goto yy727;
    goto yy725;
  yy731:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy725;
    if (yych <= 0xBF)
      goto yy729;
    goto yy725;
  yy732:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy725;
    if (yych <= 0xBF)
      goto yy729;
    goto yy725;
  yy733:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy725;
    if (yych <= 0x8F)
      goto yy729;
    goto yy725;
  yy734:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy734;
    }
    if (yych <= 0xDF) {
      if (yych <= '=') {
        if (yych <= 0x00)
          goto yy725;
        if (yych == '\n')
          goto yy725;
        goto yy723;
      } else {
        if (yych <= '>')
          goto yy736;
        if (yych <= 0x7F)
          goto yy723;
        if (yych <= 0xC1)
          goto yy725;
        goto yy727;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy728;
        if (yych == 0xED)
          goto yy730;
        goto yy729;
      } else {
        if (yych <= 0xF0)
          goto yy731;
        if (yych <= 0xF3)
          goto yy732;
        if (yych <= 0xF4)
          goto yy733;
        goto yy725;
      }
    }
  yy736:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 64) {
      goto yy723;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= '\n')
          goto yy737;
        if (yych <= ']')
          goto yy726;
      } else {
        if (yych <= 0xDF)
          goto yy727;
        if (yych <= 0xE0)
          goto yy728;
        goto yy729;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy730;
        if (yych <= 0xEF)
          goto yy729;
        goto yy731;
      } else {
        if (yych <= 0xF3)
          goto yy732;
        if (yych <= 0xF4)
          goto yy733;
      }
    }
  yy737 : { return (bufsize_t)(p - start); }
  }
}

// Try to match a link title (in single quotes, in double quotes, or
// in parentheses), returning number of chars matched.  Allow one
// level of internal nesting (quotes within quotes).
bufsize_t _scan_link_title(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    static const unsigned char yybm[] = {
        0,   208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208,
        208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208,
        208, 208, 208, 208, 208, 208, 192, 208, 208, 208, 208, 144, 80,  80,
        208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208,
        208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208,
        208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208,
        208, 208, 208, 208, 208, 208, 208, 208, 32,  208, 208, 208, 208, 208,
        208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208,
        208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208, 208,
        208, 208, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych <= '&') {
      if (yych == '"')
        goto yy742;
    } else {
      if (yych <= '\'')
        goto yy743;
      if (yych <= '(')
        goto yy744;
    }
    ++p;
  yy741 : { return 0; }
  yy742:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x00)
      goto yy741;
    if (yych <= 0x7F)
      goto yy746;
    if (yych <= 0xC1)
      goto yy741;
    if (yych <= 0xF4)
      goto yy746;
    goto yy741;
  yy743:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= 0x00)
      goto yy741;
    if (yych <= 0x7F)
      goto yy760;
    if (yych <= 0xC1)
      goto yy741;
    if (yych <= 0xF4)
      goto yy760;
    goto yy741;
  yy744:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych <= '(') {
      if (yych <= 0x00)
        goto yy741;
      if (yych <= '\'')
        goto yy773;
      goto yy741;
    } else {
      if (yych <= 0x7F)
        goto yy773;
      if (yych <= 0xC1)
        goto yy741;
      if (yych <= 0xF4)
        goto yy773;
      goto yy741;
    }
  yy745:
    yych = *++p;
  yy746:
    if (yybm[0 + yych] & 16) {
      goto yy745;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= 0x00)
          goto yy747;
        if (yych <= '"')
          goto yy748;
        goto yy750;
      } else {
        if (yych <= 0xC1)
          goto yy747;
        if (yych <= 0xDF)
          goto yy752;
        goto yy753;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy755;
        goto yy754;
      } else {
        if (yych <= 0xF0)
          goto yy756;
        if (yych <= 0xF3)
          goto yy757;
        if (yych <= 0xF4)
          goto yy758;
      }
    }
  yy747:
    p = marker;
    if (yyaccept <= 1) {
      if (yyaccept == 0) {
        goto yy741;
      } else {
        goto yy749;
      }
    } else {
      if (yyaccept == 2) {
        goto yy762;
      } else {
        goto yy775;
      }
    }
  yy748:
    ++p;
  yy749 : { return (bufsize_t)(p - start); }
  yy750:
    yych = *++p;
    if (yybm[0 + yych] & 16) {
      goto yy745;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= 0x00)
          goto yy747;
        if (yych <= '"')
          goto yy785;
        goto yy750;
      } else {
        if (yych <= 0xC1)
          goto yy747;
        if (yych >= 0xE0)
          goto yy753;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy755;
        goto yy754;
      } else {
        if (yych <= 0xF0)
          goto yy756;
        if (yych <= 0xF3)
          goto yy757;
        if (yych <= 0xF4)
          goto yy758;
        goto yy747;
      }
    }
  yy752:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy745;
    goto yy747;
  yy753:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy752;
    goto yy747;
  yy754:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy752;
    goto yy747;
  yy755:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0x9F)
      goto yy752;
    goto yy747;
  yy756:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy754;
    goto yy747;
  yy757:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy754;
    goto yy747;
  yy758:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0x8F)
      goto yy754;
    goto yy747;
  yy759:
    yych = *++p;
  yy760:
    if (yybm[0 + yych] & 64) {
      goto yy759;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= 0x00)
          goto yy747;
        if (yych >= '(')
          goto yy763;
      } else {
        if (yych <= 0xC1)
          goto yy747;
        if (yych <= 0xDF)
          goto yy765;
        goto yy766;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy768;
        goto yy767;
      } else {
        if (yych <= 0xF0)
          goto yy769;
        if (yych <= 0xF3)
          goto yy770;
        if (yych <= 0xF4)
          goto yy771;
        goto yy747;
      }
    }
  yy761:
    ++p;
  yy762 : { return (bufsize_t)(p - start); }
  yy763:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy759;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= 0x00)
          goto yy747;
        if (yych <= '\'')
          goto yy786;
        goto yy763;
      } else {
        if (yych <= 0xC1)
          goto yy747;
        if (yych >= 0xE0)
          goto yy766;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy768;
        goto yy767;
      } else {
        if (yych <= 0xF0)
          goto yy769;
        if (yych <= 0xF3)
          goto yy770;
        if (yych <= 0xF4)
          goto yy771;
        goto yy747;
      }
    }
  yy765:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy759;
    goto yy747;
  yy766:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy765;
    goto yy747;
  yy767:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy765;
    goto yy747;
  yy768:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0x9F)
      goto yy765;
    goto yy747;
  yy769:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy767;
    goto yy747;
  yy770:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy767;
    goto yy747;
  yy771:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0x8F)
      goto yy767;
    goto yy747;
  yy772:
    yych = *++p;
  yy773:
    if (yybm[0 + yych] & 128) {
      goto yy772;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= '(')
          goto yy747;
        if (yych >= '*')
          goto yy776;
      } else {
        if (yych <= 0xC1)
          goto yy747;
        if (yych <= 0xDF)
          goto yy778;
        goto yy779;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy781;
        goto yy780;
      } else {
        if (yych <= 0xF0)
          goto yy782;
        if (yych <= 0xF3)
          goto yy783;
        if (yych <= 0xF4)
          goto yy784;
        goto yy747;
      }
    }
  yy774:
    ++p;
  yy775 : { return (bufsize_t)(p - start); }
  yy776:
    yych = *++p;
    if (yych <= 0xDF) {
      if (yych <= '[') {
        if (yych <= 0x00)
          goto yy747;
        if (yych == ')')
          goto yy787;
        goto yy772;
      } else {
        if (yych <= '\\')
          goto yy776;
        if (yych <= 0x7F)
          goto yy772;
        if (yych <= 0xC1)
          goto yy747;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0)
          goto yy779;
        if (yych == 0xED)
          goto yy781;
        goto yy780;
      } else {
        if (yych <= 0xF0)
          goto yy782;
        if (yych <= 0xF3)
          goto yy783;
        if (yych <= 0xF4)
          goto yy784;
        goto yy747;
      }
    }
  yy778:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy772;
    goto yy747;
  yy779:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy778;
    goto yy747;
  yy780:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy778;
    goto yy747;
  yy781:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0x9F)
      goto yy778;
    goto yy747;
  yy782:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy780;
    goto yy747;
  yy783:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0xBF)
      goto yy780;
    goto yy747;
  yy784:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy747;
    if (yych <= 0x8F)
      goto yy780;
    goto yy747;
  yy785:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 16) {
      goto yy745;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= 0x00)
          goto yy749;
        if (yych <= '"')
          goto yy748;
        goto yy750;
      } else {
        if (yych <= 0xC1)
          goto yy749;
        if (yych <= 0xDF)
          goto yy752;
        goto yy753;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy755;
        goto yy754;
      } else {
        if (yych <= 0xF0)
          goto yy756;
        if (yych <= 0xF3)
          goto yy757;
        if (yych <= 0xF4)
          goto yy758;
        goto yy749;
      }
    }
  yy786:
    yyaccept = 2;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 64) {
      goto yy759;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= 0x00)
          goto yy762;
        if (yych <= '\'')
          goto yy761;
        goto yy763;
      } else {
        if (yych <= 0xC1)
          goto yy762;
        if (yych <= 0xDF)
          goto yy765;
        goto yy766;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy768;
        goto yy767;
      } else {
        if (yych <= 0xF0)
          goto yy769;
        if (yych <= 0xF3)
          goto yy770;
        if (yych <= 0xF4)
          goto yy771;
        goto yy762;
      }
    }
  yy787:
    yyaccept = 3;
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy772;
    }
    if (yych <= 0xE0) {
      if (yych <= '\\') {
        if (yych <= '(')
          goto yy775;
        if (yych <= ')')
          goto yy774;
        goto yy776;
      } else {
        if (yych <= 0xC1)
          goto yy775;
        if (yych <= 0xDF)
          goto yy778;
        goto yy779;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych == 0xED)
          goto yy781;
        goto yy780;
      } else {
        if (yych <= 0xF0)
          goto yy782;
        if (yych <= 0xF3)
          goto yy783;
        if (yych <= 0xF4)
          goto yy784;
        goto yy775;
      }
    }
  }
}

// Match space characters, including newlines.
bufsize_t _scan_spacechars(const unsigned char *p) {
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 128, 128, 128, 128, 128, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   128, 0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0,   0,   0,   0, 0,
    };
    yych = *p;
    if (yybm[0 + yych] & 128) {
      goto yy792;
    }
    ++p;
    { return 0; }
  yy792:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy792;
    }
    { return (bufsize_t)(p - start); }
  }
}

// Match ATX heading start.
bufsize_t _scan_atx_heading_start(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0, 128, 0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   128, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0,   0,   0, 0, 0,
    };
    yych = *p;
    if (yych == '#')
      goto yy799;
    ++p;
  yy798 : { return 0; }
  yy799:
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy800;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy798;
      if (yych <= '\n')
        goto yy803;
      goto yy798;
    } else {
      if (yych <= '\r')
        goto yy803;
      if (yych == '#')
        goto yy804;
      goto yy798;
    }
  yy800:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy800;
    }
  yy802 : { return (bufsize_t)(p - start); }
  yy803:
    ++p;
    goto yy802;
  yy804:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy800;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy805;
      if (yych <= '\n')
        goto yy803;
    } else {
      if (yych <= '\r')
        goto yy803;
      if (yych == '#')
        goto yy806;
    }
  yy805:
    p = marker;
    goto yy798;
  yy806:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy800;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy805;
      if (yych <= '\n')
        goto yy803;
      goto yy805;
    } else {
      if (yych <= '\r')
        goto yy803;
      if (yych != '#')
        goto yy805;
    }
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy800;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy805;
      if (yych <= '\n')
        goto yy803;
      goto yy805;
    } else {
      if (yych <= '\r')
        goto yy803;
      if (yych != '#')
        goto yy805;
    }
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy800;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy805;
      if (yych <= '\n')
        goto yy803;
      goto yy805;
    } else {
      if (yych <= '\r')
        goto yy803;
      if (yych != '#')
        goto yy805;
    }
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy800;
    }
    if (yych <= 0x08)
      goto yy805;
    if (yych <= '\n')
      goto yy803;
    if (yych == '\r')
      goto yy803;
    goto yy805;
  }
}

// Match setext heading line.  Return 1 for level-1 heading,
// 2 for level-2, 0 for no match.
bufsize_t _scan_setext_heading_line(const unsigned char *p) {
  const unsigned char *marker = NULL;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0, 0,  0, 0, 0, 0, 0, 0, 0, 32, 0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  32, 0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 64, 0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 128, 0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0, 0, 0, 0, 0,   0, 0, 0, 0,
        0, 0,  0, 0, 0, 0, 0, 0, 0, 0,  0,  0, 0, 0,
    };
    yych = *p;
    if (yych == '-')
      goto yy814;
    if (yych == '=')
      goto yy815;
    ++p;
  yy813 : { return 0; }
  yy814:
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 64) {
      goto yy821;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy813;
      if (yych <= '\n')
        goto yy817;
      goto yy813;
    } else {
      if (yych <= '\r')
        goto yy817;
      if (yych == ' ')
        goto yy817;
      goto yy813;
    }
  yy815:
    yych = *(marker = ++p);
    if (yybm[0 + yych] & 128) {
      goto yy827;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy813;
      if (yych <= '\n')
        goto yy824;
      goto yy813;
    } else {
      if (yych <= '\r')
        goto yy824;
      if (yych == ' ')
        goto yy824;
      goto yy813;
    }
  yy816:
    yych = *++p;
  yy817:
    if (yybm[0 + yych] & 32) {
      goto yy816;
    }
    if (yych <= 0x08)
      goto yy818;
    if (yych <= '\n')
      goto yy819;
    if (yych == '\r')
      goto yy819;
  yy818:
    p = marker;
    goto yy813;
  yy819:
    ++p;
    { return 2; }
  yy821:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy816;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy818;
      if (yych <= '\n')
        goto yy819;
      goto yy818;
    } else {
      if (yych <= '\r')
        goto yy819;
      if (yych == '-')
        goto yy821;
      goto yy818;
    }
  yy823:
    yych = *++p;
  yy824:
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy818;
      if (yych <= '\t')
        goto yy823;
      if (yych >= '\v')
        goto yy818;
    } else {
      if (yych <= '\r')
        goto yy825;
      if (yych == ' ')
        goto yy823;
      goto yy818;
    }
  yy825:
    ++p;
    { return 1; }
  yy827:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy827;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy818;
      if (yych <= '\t')
        goto yy823;
      if (yych <= '\n')
        goto yy825;
      goto yy818;
    } else {
      if (yych <= '\r')
        goto yy825;
      if (yych == ' ')
        goto yy823;
      goto yy818;
    }
  }
}

// Scan an opening code fence.
bufsize_t _scan_open_code_fence(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0,   192, 192, 192, 192, 192, 192, 192, 192, 192, 0,   192, 192, 0,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 144, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192, 192,
        224, 192, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
        0,   0,   0,   0,
    };
    yych = *p;
    if (yych == '`')
      goto yy833;
    if (yych == '~')
      goto yy834;
    ++p;
  yy832 : { return 0; }
  yy833:
    yych = *(marker = ++p);
    if (yych == '`')
      goto yy835;
    goto yy832;
  yy834:
    yych = *(marker = ++p);
    if (yych == '~')
      goto yy837;
    goto yy832;
  yy835:
    yych = *++p;
    if (yybm[0 + yych] & 16) {
      goto yy838;
    }
  yy836:
    p = marker;
    goto yy832;
  yy837:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy840;
    }
    goto yy836;
  yy838:
    yych = *++p;
    if (yybm[0 + yych] & 16) {
      goto yy838;
    }
    if (yych <= 0xDF) {
      if (yych <= '\f') {
        if (yych <= 0x00)
          goto yy836;
        if (yych == '\n') {
          marker = p;
          goto yy844;
        }
        marker = p;
        goto yy842;
      } else {
        if (yych <= '\r') {
          marker = p;
          goto yy844;
        }
        if (yych <= 0x7F) {
          marker = p;
          goto yy842;
        }
        if (yych <= 0xC1)
          goto yy836;
        marker = p;
        goto yy846;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0) {
          marker = p;
          goto yy847;
        }
        if (yych == 0xED) {
          marker = p;
          goto yy849;
        }
        marker = p;
        goto yy848;
      } else {
        if (yych <= 0xF0) {
          marker = p;
          goto yy850;
        }
        if (yych <= 0xF3) {
          marker = p;
          goto yy851;
        }
        if (yych <= 0xF4) {
          marker = p;
          goto yy852;
        }
        goto yy836;
      }
    }
  yy840:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy840;
    }
    if (yych <= 0xDF) {
      if (yych <= '\f') {
        if (yych <= 0x00)
          goto yy836;
        if (yych == '\n') {
          marker = p;
          goto yy855;
        }
        marker = p;
        goto yy853;
      } else {
        if (yych <= '\r') {
          marker = p;
          goto yy855;
        }
        if (yych <= 0x7F) {
          marker = p;
          goto yy853;
        }
        if (yych <= 0xC1)
          goto yy836;
        marker = p;
        goto yy857;
      }
    } else {
      if (yych <= 0xEF) {
        if (yych <= 0xE0) {
          marker = p;
          goto yy858;
        }
        if (yych == 0xED) {
          marker = p;
          goto yy860;
        }
        marker = p;
        goto yy859;
      } else {
        if (yych <= 0xF0) {
          marker = p;
          goto yy861;
        }
        if (yych <= 0xF3) {
          marker = p;
          goto yy862;
        }
        if (yych <= 0xF4) {
          marker = p;
          goto yy863;
        }
        goto yy836;
      }
    }
  yy842:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy842;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy836;
        if (yych >= 0x0E)
          goto yy836;
      } else {
        if (yych <= 0xDF)
          goto yy846;
        if (yych <= 0xE0)
          goto yy847;
        goto yy848;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy849;
        if (yych <= 0xEF)
          goto yy848;
        goto yy850;
      } else {
        if (yych <= 0xF3)
          goto yy851;
        if (yych <= 0xF4)
          goto yy852;
        goto yy836;
      }
    }
  yy844:
    ++p;
    p = marker;
    { return (bufsize_t)(p - start); }
  yy846:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy842;
    goto yy836;
  yy847:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy846;
    goto yy836;
  yy848:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy846;
    goto yy836;
  yy849:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0x9F)
      goto yy846;
    goto yy836;
  yy850:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy848;
    goto yy836;
  yy851:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy848;
    goto yy836;
  yy852:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0x8F)
      goto yy848;
    goto yy836;
  yy853:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy853;
    }
    if (yych <= 0xEC) {
      if (yych <= 0xC1) {
        if (yych <= 0x00)
          goto yy836;
        if (yych >= 0x0E)
          goto yy836;
      } else {
        if (yych <= 0xDF)
          goto yy857;
        if (yych <= 0xE0)
          goto yy858;
        goto yy859;
      }
    } else {
      if (yych <= 0xF0) {
        if (yych <= 0xED)
          goto yy860;
        if (yych <= 0xEF)
          goto yy859;
        goto yy861;
      } else {
        if (yych <= 0xF3)
          goto yy862;
        if (yych <= 0xF4)
          goto yy863;
        goto yy836;
      }
    }
  yy855:
    ++p;
    p = marker;
    { return (bufsize_t)(p - start); }
  yy857:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy853;
    goto yy836;
  yy858:
    yych = *++p;
    if (yych <= 0x9F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy857;
    goto yy836;
  yy859:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy857;
    goto yy836;
  yy860:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0x9F)
      goto yy857;
    goto yy836;
  yy861:
    yych = *++p;
    if (yych <= 0x8F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy859;
    goto yy836;
  yy862:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0xBF)
      goto yy859;
    goto yy836;
  yy863:
    yych = *++p;
    if (yych <= 0x7F)
      goto yy836;
    if (yych <= 0x8F)
      goto yy859;
    goto yy836;
  }
}

// Scan a closing code fence with length at least len.
bufsize_t _scan_close_code_fence(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    static const unsigned char yybm[] = {
        0, 0, 0, 0, 0, 0, 0, 0, 0,  128, 0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   128, 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 32, 0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 64, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0, 0, 0, 0,  0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0,  0,   0,   0, 0, 0,
    };
    yych = *p;
    if (yych == '`')
      goto yy868;
    if (yych == '~')
      goto yy869;
    ++p;
  yy867 : { return 0; }
  yy868:
    yych = *(marker = ++p);
    if (yych == '`')
      goto yy870;
    goto yy867;
  yy869:
    yych = *(marker = ++p);
    if (yych == '~')
      goto yy872;
    goto yy867;
  yy870:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy873;
    }
  yy871:
    p = marker;
    goto yy867;
  yy872:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy875;
    }
    goto yy871;
  yy873:
    yych = *++p;
    if (yybm[0 + yych] & 32) {
      goto yy873;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy871;
      if (yych <= '\t') {
        marker = p;
        goto yy877;
      }
      if (yych <= '\n') {
        marker = p;
        goto yy879;
      }
      goto yy871;
    } else {
      if (yych <= '\r') {
        marker = p;
        goto yy879;
      }
      if (yych == ' ') {
        marker = p;
        goto yy877;
      }
      goto yy871;
    }
  yy875:
    yych = *++p;
    if (yybm[0 + yych] & 64) {
      goto yy875;
    }
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy871;
      if (yych <= '\t') {
        marker = p;
        goto yy881;
      }
      if (yych <= '\n') {
        marker = p;
        goto yy883;
      }
      goto yy871;
    } else {
      if (yych <= '\r') {
        marker = p;
        goto yy883;
      }
      if (yych == ' ') {
        marker = p;
        goto yy881;
      }
      goto yy871;
    }
  yy877:
    yych = *++p;
    if (yybm[0 + yych] & 128) {
      goto yy877;
    }
    if (yych <= 0x08)
      goto yy871;
    if (yych <= '\n')
      goto yy879;
    if (yych != '\r')
      goto yy871;
  yy879:
    ++p;
    p = marker;
    { return (bufsize_t)(p - start); }
  yy881:
    yych = *++p;
    if (yych <= '\f') {
      if (yych <= 0x08)
        goto yy871;
      if (yych <= '\t')
        goto yy881;
      if (yych >= '\v')
        goto yy871;
    } else {
      if (yych <= '\r')
        goto yy883;
      if (yych == ' ')
        goto yy881;
      goto yy871;
    }
  yy883:
    ++p;
    p = marker;
    { return (bufsize_t)(p - start); }
  }
}

// Scans an entity.
// Returns number of chars matched.
bufsize_t _scan_entity(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    yych = *p;
    if (yych == '&')
      goto yy889;
    ++p;
  yy888 : { return 0; }
  yy889:
    yych = *(marker = ++p);
    if (yych <= '@') {
      if (yych != '#')
        goto yy888;
    } else {
      if (yych <= 'Z')
        goto yy892;
      if (yych <= '`')
        goto yy888;
      if (yych <= 'z')
        goto yy892;
      goto yy888;
    }
    yych = *++p;
    if (yych <= 'W') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy893;
    } else {
      if (yych <= 'X')
        goto yy894;
      if (yych == 'x')
        goto yy894;
    }
  yy891:
    p = marker;
    goto yy888;
  yy892:
    yych = *++p;
    if (yych <= '@') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy895;
      goto yy891;
    } else {
      if (yych <= 'Z')
        goto yy895;
      if (yych <= '`')
        goto yy891;
      if (yych <= 'z')
        goto yy895;
      goto yy891;
    }
  yy893:
    yych = *++p;
    if (yych <= '/')
      goto yy891;
    if (yych <= '9')
      goto yy896;
    if (yych == ';')
      goto yy897;
    goto yy891;
  yy894:
    yych = *++p;
    if (yych <= '@') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy899;
      goto yy891;
    } else {
      if (yych <= 'F')
        goto yy899;
      if (yych <= '`')
        goto yy891;
      if (yych <= 'f')
        goto yy899;
      goto yy891;
    }
  yy895:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy900;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
        goto yy900;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'z')
          goto yy900;
        goto yy891;
      }
    }
  yy896:
    yych = *++p;
    if (yych <= '/')
      goto yy891;
    if (yych <= '9')
      goto yy901;
    if (yych != ';')
      goto yy891;
  yy897:
    ++p;
    { return (bufsize_t)(p - start); }
  yy899:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy902;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'F') {
        if (yych <= '@')
          goto yy891;
        goto yy902;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'f')
          goto yy902;
        goto yy891;
      }
    }
  yy900:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy903;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
        goto yy903;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'z')
          goto yy903;
        goto yy891;
      }
    }
  yy901:
    yych = *++p;
    if (yych <= '/')
      goto yy891;
    if (yych <= '9')
      goto yy904;
    if (yych == ';')
      goto yy897;
    goto yy891;
  yy902:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy905;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'F') {
        if (yych <= '@')
          goto yy891;
        goto yy905;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'f')
          goto yy905;
        goto yy891;
      }
    }
  yy903:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy906;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
        goto yy906;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'z')
          goto yy906;
        goto yy891;
      }
    }
  yy904:
    yych = *++p;
    if (yych <= '/')
      goto yy891;
    if (yych <= '9')
      goto yy907;
    if (yych == ';')
      goto yy897;
    goto yy891;
  yy905:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy908;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'F') {
        if (yych <= '@')
          goto yy891;
        goto yy908;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'f')
          goto yy908;
        goto yy891;
      }
    }
  yy906:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy909;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
        goto yy909;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'z')
          goto yy909;
        goto yy891;
      }
    }
  yy907:
    yych = *++p;
    if (yych <= '/')
      goto yy891;
    if (yych <= '9')
      goto yy910;
    if (yych == ';')
      goto yy897;
    goto yy891;
  yy908:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy911;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'F') {
        if (yych <= '@')
          goto yy891;
        goto yy911;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'f')
          goto yy911;
        goto yy891;
      }
    }
  yy909:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy912;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
        goto yy912;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'z')
          goto yy912;
        goto yy891;
      }
    }
  yy910:
    yych = *++p;
    if (yych <= '/')
      goto yy891;
    if (yych <= '9')
      goto yy913;
    if (yych == ';')
      goto yy897;
    goto yy891;
  yy911:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy913;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'F') {
        if (yych <= '@')
          goto yy891;
        goto yy913;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'f')
          goto yy913;
        goto yy891;
      }
    }
  yy912:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy914;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
        goto yy914;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'z')
          goto yy914;
        goto yy891;
      }
    }
  yy913:
    yych = *++p;
    if (yych == ';')
      goto yy897;
    goto yy891;
  yy914:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy915;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy915:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy916;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy916:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy917;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy917:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy918;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy918:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy919;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy919:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy920;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy920:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy921;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy921:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy922;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy922:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy923;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy923:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy924;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy924:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy925;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy925:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy926;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy926:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy927;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy927:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy928;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy928:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy929;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy929:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy930;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy930:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy931;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy931:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy932;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy932:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy933;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy933:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy934;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy934:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy935;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy935:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy936;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy936:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy937;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych >= '{')
          goto yy891;
      }
    }
  yy937:
    yych = *++p;
    if (yych <= ';') {
      if (yych <= '/')
        goto yy891;
      if (yych <= '9')
        goto yy913;
      if (yych <= ':')
        goto yy891;
      goto yy897;
    } else {
      if (yych <= 'Z') {
        if (yych <= '@')
          goto yy891;
        goto yy913;
      } else {
        if (yych <= '`')
          goto yy891;
        if (yych <= 'z')
          goto yy913;
        goto yy891;
      }
    }
  }
}

// Returns positive value if a URL begins in a way that is potentially
// dangerous, with javascript:, vbscript:, file:, or data:, otherwise 0.
bufsize_t _scan_dangerous_url(const unsigned char *p) {
  const unsigned char *marker = NULL;
  const unsigned char *start = p;

  {
    unsigned char yych;
    unsigned int yyaccept = 0;
    yych = *p;
    if (yych <= 'V') {
      if (yych <= 'F') {
        if (yych == 'D')
          goto yy942;
        if (yych >= 'F')
          goto yy943;
      } else {
        if (yych == 'J')
          goto yy944;
        if (yych >= 'V')
          goto yy945;
      }
    } else {
      if (yych <= 'f') {
        if (yych == 'd')
          goto yy942;
        if (yych >= 'f')
          goto yy943;
      } else {
        if (yych <= 'j') {
          if (yych >= 'j')
            goto yy944;
        } else {
          if (yych == 'v')
            goto yy945;
        }
      }
    }
    ++p;
  yy941 : { return 0; }
  yy942:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych == 'A')
      goto yy946;
    if (yych == 'a')
      goto yy946;
    goto yy941;
  yy943:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych == 'I')
      goto yy948;
    if (yych == 'i')
      goto yy948;
    goto yy941;
  yy944:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych == 'A')
      goto yy949;
    if (yych == 'a')
      goto yy949;
    goto yy941;
  yy945:
    yyaccept = 0;
    yych = *(marker = ++p);
    if (yych == 'B')
      goto yy950;
    if (yych == 'b')
      goto yy950;
    goto yy941;
  yy946:
    yych = *++p;
    if (yych == 'T')
      goto yy951;
    if (yych == 't')
      goto yy951;
  yy947:
    p = marker;
    if (yyaccept == 0) {
      goto yy941;
    } else {
      goto yy959;
    }
  yy948:
    yych = *++p;
    if (yych == 'L')
      goto yy952;
    if (yych == 'l')
      goto yy952;
    goto yy947;
  yy949:
    yych = *++p;
    if (yych == 'V')
      goto yy953;
    if (yych == 'v')
      goto yy953;
    goto yy947;
  yy950:
    yych = *++p;
    if (yych == 'S')
      goto yy954;
    if (yych == 's')
      goto yy954;
    goto yy947;
  yy951:
    yych = *++p;
    if (yych == 'A')
      goto yy955;
    if (yych == 'a')
      goto yy955;
    goto yy947;
  yy952:
    yych = *++p;
    if (yych == 'E')
      goto yy956;
    if (yych == 'e')
      goto yy956;
    goto yy947;
  yy953:
    yych = *++p;
    if (yych == 'A')
      goto yy950;
    if (yych == 'a')
      goto yy950;
    goto yy947;
  yy954:
    yych = *++p;
    if (yych == 'C')
      goto yy957;
    if (yych == 'c')
      goto yy957;
    goto yy947;
  yy955:
    yych = *++p;
    if (yych == ':')
      goto yy958;
    goto yy947;
  yy956:
    yych = *++p;
    if (yych == ':')
      goto yy960;
    goto yy947;
  yy957:
    yych = *++p;
    if (yych == 'R')
      goto yy961;
    if (yych == 'r')
      goto yy961;
    goto yy947;
  yy958:
    yyaccept = 1;
    yych = *(marker = ++p);
    if (yych == 'I')
      goto yy962;
    if (yych == 'i')
      goto yy962;
  yy959 : { return (bufsize_t)(p - start); }
  yy960:
    ++p;
    goto yy959;
  yy961:
    yych = *++p;
    if (yych == 'I')
      goto yy963;
    if (yych == 'i')
      goto yy963;
    goto yy947;
  yy962:
    yych = *++p;
    if (yych == 'M')
      goto yy964;
    if (yych == 'm')
      goto yy964;
    goto yy947;
  yy963:
    yych = *++p;
    if (yych == 'P')
      goto yy965;
    if (yych == 'p')
      goto yy965;
    goto yy947;
  yy964:
    yych = *++p;
    if (yych == 'A')
      goto yy966;
    if (yych == 'a')
      goto yy966;
    goto yy947;
  yy965:
    yych = *++p;
    if (yych == 'T')
      goto yy956;
    if (yych == 't')
      goto yy956;
    goto yy947;
  yy966:
    yych = *++p;
    if (yych == 'G')
      goto yy967;
    if (yych != 'g')
      goto yy947;
  yy967:
    yych = *++p;
    if (yych == 'E')
      goto yy968;
    if (yych != 'e')
      goto yy947;
  yy968:
    yych = *++p;
    if (yych != '/')
      goto yy947;
    yych = *++p;
    if (yych <= 'W') {
      if (yych <= 'J') {
        if (yych == 'G')
          goto yy970;
        if (yych <= 'I')
          goto yy947;
        goto yy971;
      } else {
        if (yych == 'P')
          goto yy972;
        if (yych <= 'V')
          goto yy947;
        goto yy973;
      }
    } else {
      if (yych <= 'j') {
        if (yych == 'g')
          goto yy970;
        if (yych <= 'i')
          goto yy947;
        goto yy971;
      } else {
        if (yych <= 'p') {
          if (yych <= 'o')
            goto yy947;
          goto yy972;
        } else {
          if (yych == 'w')
            goto yy973;
          goto yy947;
        }
      }
    }
  yy970:
    yych = *++p;
    if (yych == 'I')
      goto yy974;
    if (yych == 'i')
      goto yy974;
    goto yy947;
  yy971:
    yych = *++p;
    if (yych == 'P')
      goto yy975;
    if (yych == 'p')
      goto yy975;
    goto yy947;
  yy972:
    yych = *++p;
    if (yych == 'N')
      goto yy976;
    if (yych == 'n')
      goto yy976;
    goto yy947;
  yy973:
    yych = *++p;
    if (yych == 'E')
      goto yy977;
    if (yych == 'e')
      goto yy977;
    goto yy947;
  yy974:
    yych = *++p;
    if (yych == 'F')
      goto yy978;
    if (yych == 'f')
      goto yy978;
    goto yy947;
  yy975:
    yych = *++p;
    if (yych == 'E')
      goto yy976;
    if (yych != 'e')
      goto yy947;
  yy976:
    yych = *++p;
    if (yych == 'G')
      goto yy978;
    if (yych == 'g')
      goto yy978;
    goto yy947;
  yy977:
    yych = *++p;
    if (yych == 'B')
      goto yy980;
    if (yych == 'b')
      goto yy980;
    goto yy947;
  yy978:
    ++p;
    { return 0; }
  yy980:
    yych = *++p;
    if (yych == 'P')
      goto yy978;
    if (yych == 'p')
      goto yy978;
    goto yy947;
  }
}
